"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""


import traceback

import numpy as np
import sympy


def bin_rep(n, bits):
    if n < 0:
        raise Exception("Only positive numbers are supported")

    if n >= 2**bits:
        raise Exception(
            str(n) + " can't be represented as a " + str(bits) + " bit number"
        )

    zero_string = "".join(["0" for k in range(bits)])
    return (zero_string + bin(n)[2:])[-bits:]


def int_encoder(qv, encoding_number):
    if encoding_number > 2 ** len(qv) - 1:
        raise Exception("Not enough qubits to encode integer " + str(encoding_number))

    binary_rep = bin_rep(encoding_number, len(qv))[::-1]
    from qrisp import x

    for i in range(len(binary_rep)):
        if int(binary_rep[i]):
            x(qv[i])


# Calculates the binary expression of a given integer and returns it as an array of
# length bits
def int_as_array(k, bit):
    bin_str = bin_rep(k, bit)
    return np.array([int(c) for c in bin_str])


def array_as_int(array):
    result = 0
    for k in range(len(array)):
        if array[::-1][k]:
            result += 2 ** (k)

    return result


# Decomposes the circuit qc until no more decompositions are possible and then counts
# the cnot operations
def cnot_count(qc):
    qc = qc.transpile()

    gate_count_dic = qc.count_ops()

    try:
        return gate_count_dic["cx"]
    except KeyError:
        return 0


def is_inv(x, bit):
    # return (math.gcd(int(np.round(x, 3)),2**bit) == 1)

    # The only divisors 2**bit has is powers of 2
    # ie. if tha factorization of x doesn't contain any powers of 2 it is invertible
    # in other words: x is invertible if it is uneven
    return bool(int(x) % 2)


def get_depth_dic(qc, transpile_qc=True):
    if len(qc.qubits) == 0:
        return {}

    if transpile_qc:
        qc = qc.transpile()

    # Assign each bit in the circuit a unique integer
    # to index into op_stack.
    bit_indices = {bit: idx for idx, bit in enumerate(qc.qubits + qc.clbits)}

    # If no bits, return 0
    if not bit_indices:
        return 0

    # A list that holds the height of each qubit
    # and classical bit.
    op_stack = [0] * len(bit_indices)

    # Here we are playing a modified version of
    # Tetris where we stack gates, but multi-qubit
    # gates, or measurements have a block for each
    # qubit or cbit that are connected by a virtual
    # line so that they all stacked at the same depth.
    # Conditional gates act on all cbits in the register
    # they are conditioned on.
    # We treat barriers or snapshots different as
    # They are transpiler and simulator directives.
    # The max stack height is the circuit depth.

    for instr in qc.data:
        if instr.op.name in ["qb_alloc", "qb_dealloc"]:
            continue
        qargs = instr.qubits
        cargs = instr.clbits

        levels = []
        reg_ints = []
        # If count then add one to stack heights
        count = True

        for ind, reg in enumerate(qargs + cargs):
            # Add to the stacks of the qubits and
            # cbits used in the gate.
            reg_ints.append(bit_indices[reg])
            if count:
                levels.append(op_stack[reg_ints[ind]] + 1)
            else:
                levels.append(op_stack[reg_ints[ind]])

        max_level = max(levels)
        for ind in reg_ints:
            op_stack[ind] = max_level

    return {qc.qubits[i]: op_stack[i] for i in range(len(qc.qubits))}


def gate_wrap(*args, permeability=None, is_qfree=None, name=None, verify=False):
    """
    Decorator to bundle up the quantum instructions of a function into a single gate
    object. Bundled gate objects can help debugging as it allows for a more clear
    QuantumCircuit visualisation.

    Furthermore, bundling up functions is relevant for Qrisps uncomputation algorithm.
    When bundling up for uncomputation, this decorator provides the means to annotate
    the gate objects with information about its permeability and qfree-ness. For further
    information about these concepts check the
    :ref:`uncomputation documentation<uncomputation>`. Specifying this information
    allows to skip the computationally costly automatic determination at runtime.

    Note that the specified information is not checked for correctness as this would
    defy the purpose.

    A shorthand for ``gate_wrap(permeability = "args", is_qfree = True)`` is the
    :meth:`lifted <qrisp.lifted>` decorator.


    .. warning::

        Using ``gate_wrap`` without specifying permeability and ``qfree``-ness on
        functions processing a lot of qubits, can causes long compile times, since the
        unitaries of these gates have to be determined numerically.

    .. warning::

        Incorrect information about permeability and ``qfree``-ness can yield incorrect
        compilation results. If you are unsure, use the ``verify`` keyword on a small
        scale first.


    Parameters
    ----------

    permeability : string or list, optional
        Specify the permeability behavior of the function. When given "args", it is
        assumed that the gate is permeable only on the qubits of the arguments. When
        given "full", it is assumed that the gate is permeable on every qubit it acts on
        (i.e. also the result). When given a list of integers it is assumed, that the
        gate is permeable on the qubits of the arguments corresponding to the integers.
        The default is None.
    is_qfree : bool, optional
        Specify the qfree-ness of the function. The default is None.
    name : string, optional
        String which will be used for naming the gate object. The default is None.
    verify : bool, optional
        If set to ``True``, the specified information about permeability and
        ``qfree``-ness will be checked numerically. The default is ``False``.

    Examples
    --------

    We create a simple function wrapping up multiple gates: ::


        from qrisp import QuantumVariable, cx, x, h, z, gate_wrap

        @gate_wrap
        def example_function(a, b):

            cx(a,b)
            x(a)
            cx(b,a)
            h(b)

        a = QuantumVariable(3)
        b = QuantumVariable(3)

        example_function(a, b)

    >>> print(a.qs)
    QuantumCircuit:
    ---------------
         ┌───────────────────┐
    b.0: ┤0                  ├
         │                   │
    b.1: ┤1                  ├
         │                   │
    b.2: ┤2                  ├
         │  example_function │
    a.0: ┤3                  ├
         │                   │
    a.1: ┤4                  ├
         │                   │
    a.2: ┤5                  ├
         └───────────────────┘
    Live QuantumVariables:
    ----------------------
    QuantumVariable a
    QuantumVariable b
    >>> print(a.qs.transpile())
         ┌───┐                    ┌───┐
    b.0: ┤ X ├─────────────────■──┤ H ├──────────
         └─┬─┘┌───┐            │  └───┘┌───┐
    b.1: ──┼──┤ X ├────────────┼────■──┤ H ├─────
           │  └─┬─┘┌───┐       │    │  └───┘┌───┐
    b.2: ──┼────┼──┤ X ├───────┼────┼────■──┤ H ├
           │    │  └─┬─┘┌───┐┌─┴─┐  │    │  └───┘
    a.0: ──■────┼────┼──┤ X ├┤ X ├──┼────┼───────
                │    │  ├───┤└───┘┌─┴─┐  │
    a.1: ───────■────┼──┤ X ├─────┤ X ├──┼───────
                     │  ├───┤     └───┘┌─┴─┐
    a.2: ────────────■──┤ X ├──────────┤ X ├─────
                        └───┘          └───┘


    In the next example, we create a function that performs no quantum gates and specify
    that it is qfree and permeable on the second argument but not on the first. ::

        from qrisp import QuantumCircuit

        @gate_wrap(permeability = [1], is_qfree = True)
        def example_function(arg_0, arg_1):

            res = QuantumVariable(1)

            #Append an identity gate
            res.qs.append(QuantumCircuit(3).to_gate(), [arg_0, arg_1, res])

            return res

        qv_0 = QuantumVariable(1)
        qv_1 = QuantumVariable(1)

        res = example_function(qv_0, qv_1)

    >>> print(qv_0.qs)
    QuantumCircuit:
    ---------------
            ┌───────────────────┐
    qv_0.0: ┤0                  ├
            │                   │
    qv_1.0: ┤1 example_function ├
            │                   │
     res.0: ┤2                  ├
            └───────────────────┘
    Live QuantumVariables:
    ----------------------
    QuantumVariable qv_0
    QuantumVariable qv_1
    QuantumVariable res
    >>> qv_1.uncompute()
    >>> print(qv_0.qs)
    QuantumCircuit:
    ---------------
            ┌───────────────────┐
    qv_0.0: ┤0                  ├
            │                   │
    qv_1.0: ┤1 example_function ├
            │                   │
     res.0: ┤2                  ├
            └───────────────────┘
    Live QuantumVariables:
    ----------------------
    QuantumVariable qv_0
    QuantumVariable res

    Since ``arg_1`` is marked as permeable, there are no further gates required for
    uncomputation. The situation is different for the other two QuantumVariables, where
    #the qubits are not marked as permeable.

    >>> qv_0.uncompute(do_it = False)
    >>> res.uncompute()
    >>> print(qv_0.qs)
    QuantumCircuit:
    ---------------
            ┌───────────────────┐┌──────────────────────┐
    qv_0.0: ┤0                  ├┤0                     ├
            │                   ││                      │
    qv_1.0: ┤1 example_function ├┤1 example_function_dg ├
            │                   ││                      │
     res.0: ┤2                  ├┤2                     ├
            └───────────────────┘└──────────────────────┘
    Live QuantumVariables:
    ----------------------
    """

    if len(args):
        return gate_wrap_inner(args[0])

    else:

        def gate_wrap_helper(function):
            return gate_wrap_inner(
                function,
                permeability=permeability,
                is_qfree=is_qfree,
                name=name,
                verify=verify,
            )

        return gate_wrap_helper


def gate_wrap_inner(
    function, permeability=None, is_qfree=None, name=None, verify=False
):
    def wrapped_function(
        *args, permeability=permeability, is_qfree=is_qfree, verify=verify, **kwargs
    ):
        wrapped_function.__name__ = function.__name__
        from qrisp.circuit import Qubit
        from qrisp.core import recursive_qs_search, recursive_qv_search
        from qrisp.environments import GateWrapEnvironment
        from qrisp import merge

        qs_list = recursive_qs_search([args, kwargs])

        if len(qs_list) == 0:
            raise Exception(
                "Could not find QuantumSession object in function arguments during "
                "gate-wrapping"
            )

        qs = qs_list[0]

        initial_qubits = set(qs.qubits)


        if name is None:
            gwe = GateWrapEnvironment(name=function.__name__)
        else:
            gwe = GateWrapEnvironment(name=name)

        with gwe:
            result = function(*args, **kwargs)

        if len(qs.env_stack):
            gwe.compile()
            
        if gwe in qs.data:
            qs.data.remove(gwe)
        
        if gwe.instruction is None:
            return result

        created_qubits = set(qs_list[0].qubits) - initial_qubits

        ancillas = []

        for qb in created_qubits:
            if qb.allocated is False:
                ancillas.append(qb)

        if is_qfree is not None:
            if verify and is_qfree:
                from qrisp.uncomputation import is_qfree as is_qfree_function

                if not is_qfree_function(gwe.instruction.op):
                    raise Exception(
                        f"Verification of qfree-ness for function {function.__name__} "
                        f"failed"
                    )

            gwe.instruction.op.is_qfree = is_qfree

        if permeability is not None:
            permeability_dict = {}
            arg_qubits = []

            if isinstance(permeability, list):
                args = [args[i] for i in permeability]
                permeability = "args"

            for arg in args:
                if isinstance(arg, list):
                    for item in arg:
                        if isinstance(item, Qubit):
                            arg_qubits.append(item)
                for arg_qv in recursive_qv_search([arg]):
                    arg_qubits += arg_qv.reg

            for i in range(len(gwe.instruction.qubits)):
                qb = gwe.instruction.qubits[i]

                if qb in arg_qubits:
                    if permeability in ["args", "full"]:
                        permeability_dict[i] = True
                    else:
                        raise Exception(
                            "Don't know permeability option " + permeability
                        )
                else:
                    if permeability == "full":
                        permeability_dict[i] = True
                    else:
                        permeability_dict[i] = False

            for i in range(len(gwe.instruction.qubits)):
                qb = gwe.instruction.qubits[i]

                if qb in ancillas:
                    # Even though ancilla qubits are permeable, we want to be able to
                    # use the gate_wrap decorator as an interface to perform
                    # recomputation. If we mark them as permeable, Unqomp won't  wrap
                    # the uncomputed gate in alloc/dealloc gates but instead wrap both
                    # the computation gate and the recomputation in alloc/dealloc gates

                    # To undestand this behavior better consider the following example

                    # from qrisp import QuantumFloat
                    # a = QuantumFloat(2)
                    # qf_res = a * a
                    # qf_res.uncompute()
                    # print(qf_res.qs)

                    # If we mark the ancilla qubits as permeable, this gives

                    # QuantumCircuit:
                    # ---------------
                    #              ┌──────────┐┌──────────┐┌─────────────┐
                    #         a.0: ┤ qb_alloc ├┤0         ├┤0            ├──────────────
                    #              ├──────────┤│          ││             │
                    #         a.1: ┤ qb_alloc ├┤1         ├┤1            ├──────────────
                    #              ├──────────┤│          ││             │┌────────────┐
                    #    return.0: ┤ qb_alloc ├┤2         ├┤2            ├┤ qb_dealloc ├
                    #              ├──────────┤│          ││             │├────────────┤
                    #    return.1: ┤ qb_alloc ├┤3 __mul__ ├┤3 __mul___dg ├┤ qb_dealloc ├
                    #              ├──────────┤│          ││             │├────────────┤
                    #    return.2: ┤ qb_alloc ├┤4         ├┤4            ├┤ qb_dealloc ├
                    #              ├──────────┤│          ││             │├────────────┤
                    #    return.3: ┤ qb_alloc ├┤5         ├┤5            ├┤ qb_dealloc ├
                    #              ├──────────┤│          ││             │├────────────┤
                    # sbp_anc_0.0: ┤ qb_alloc ├┤6         ├┤6            ├┤ qb_dealloc ├
                    #              └──────────┘└──────────┘└─────────────┘└────────────┘
                    # Live QuantumVariables:
                    # ----------------------
                    # QuantumFloat a

                    # If we set them to non-permeable, we get instead

                    # QuantumCircuit:
                    # ---------------
                    #              ┌──────────┐┌──────────┐                          ┌─────────────┐»
                    #         a.0: ┤ qb_alloc ├┤0         ├──────────────────────────┤0            ├»
                    #              ├──────────┤│          │                          │             │»
                    #         a.1: ┤ qb_alloc ├┤1         ├──────────────────────────┤1            ├»
                    #              ├──────────┤│          │                          │             │»
                    #    return.0: ┤ qb_alloc ├┤2         ├──────────────────────────┤2            ├»
                    #              ├──────────┤│          │                          │             │»
                    #    return.1: ┤ qb_alloc ├┤3 __mul__ ├──────────────────────────┤3 __mul___dg ├»
                    #              ├──────────┤│          │                          │             │»
                    #    return.2: ┤ qb_alloc ├┤4         ├──────────────────────────┤4            ├»
                    #              ├──────────┤│          │                          │             │»
                    #    return.3: ┤ qb_alloc ├┤5         ├──────────────────────────┤5            ├»
                    #              ├──────────┤│          │┌────────────┐┌──────────┐│             │»
                    # sbp_anc_0.0: ┤ qb_alloc ├┤6         ├┤ qb_dealloc ├┤ qb_alloc ├┤6            ├»
                    #              └──────────┘└──────────┘└────────────┘└──────────┘└─────────────┘»
                    # «
                    # «        a.0: ──────────────
                    # «
                    # «        a.1: ──────────────
                    # «             ┌────────────┐
                    # «   return.0: ┤ qb_dealloc ├
                    # «             ├────────────┤
                    # «   return.1: ┤ qb_dealloc ├
                    # «             ├────────────┤
                    # «   return.2: ┤ qb_dealloc ├
                    # «             ├────────────┤
                    # «   return.3: ┤ qb_dealloc ├
                    # «             ├────────────┤
                    # «sbp_anc_0.0: ┤ qb_dealloc ├
                    # «             └────────────┘
                    # Live QuantumVariables:
                    # ----------------------
                    # QuantumFloat a

                    # In both cases, sbp_anc is recomputed, but in the second case, the
                    # reallocation allows the compiler to use the free qubit elsewhere
                    # and afterwards pick a potentially different qubit for the
                    # recomputation.

                    permeability_dict[i] = False

                for i in range(len(gwe.instruction.qubits)):
                    if permeability_dict[i] is None:
                        permeability_dict[i] = False

                if verify:
                    from qrisp.uncomputation import is_permeable

                    permeable_qubit_indices = []

                    for i in range(gwe.instruction.op.num_qubits):
                        if permeability_dict[i]:
                            permeable_qubit_indices.append(i)

                    if not is_permeable(gwe.instruction.op, permeable_qubit_indices):
                        raise Exception(
                            f"Verification of permeability for function "
                            f"{function.__name__} failed"
                        )

                gwe.instruction.op.permeability = permeability_dict

        return result

    return wrapped_function


# Function to measure multiple quantum variables at once to assess their entanglement
def multi_measurement(qv_list, shots=10000, backend=None):
    """
    This functions facilitates the measurement of multiple QuantumVariables at the same
    time. This can be used if the entanglement structure between several
    QuantumVariables is of interest.

    Parameters
    ----------
    qv_list : list[QuantumVariable]
        A list of QuantumVariables.
    shots : int, optional
        The amount of shots to perform. The default is 10000.
    backend : BackendClient, optional
        The backend to evaluate the compiled QuantumCircuit on. By default, the backend
        from default_backend.py will be used.

    Raises
    ------
    Exception
        Tried to perform measurement with open environments.

    Returns
    -------
    counts_list : list
        A list of tuples. The first element of each tuple is a tuple again and contains
        the labels of the QuantumVariables. The second element is a float and indicates
        the probability of measurement.

    Examples
    --------

    We entangle three QuantumFloats via addition and perform a multi-measurement:

    >>> from qrisp import QuantumFloat, h, multi_measurement
    >>> qf_0 = QuantumFloat(4)
    >>> qf_1 = QuantumFloat(4)
    >>> qf_0[:] = 3
    >>> qf_1[:] = 2
    >>> h(qf_1[0])
    >>> qf_sum = qf_0 + qf_1
    >>> multi_measurement([qf_0, qf_1, qf_sum])
    {(3, 2, 5): 0.5, (3, 3, 6): 0.5}

    """

    if backend is None:
        if qv_list[0].qs.backend is None:
            from qrisp.default_backend import def_backend

            backend = def_backend
        else:
            backend = qv_list[0].qs.backend

    if len(qv_list[0].qs.env_stack) != 0:
        raise Exception("Tried to perform measurement with open environments")

    from qrisp import merge

    merge(qv_list)

    # Copy circuit in order to prevent modification
    from qrisp import QuantumArray, QuantumVariable, recursive_qv_search
    from qrisp.core.compilation import qompiler

    temp = recursive_qv_search(qv_list)

    compiled_qc = qompiler(
        qv_list[0].qs, intended_measurements=sum([qv.reg for qv in temp], [])
    )
    # compiled_qc = qv_list[0].qs.copy()
    # Add classical registers for the measurement results to be stored in
    cl_reg_list = []

    for var in qv_list[::-1]:
        cl_reg = []

        if isinstance(var, QuantumArray):
            qubits = sum([qv.reg for qv in var.flatten()[::-1]], [])
        elif isinstance(var, QuantumVariable):
            qubits = var.reg
        else:
            raise Exception(f"Found type {type(var)} in measurement list")

        for i in range(len(qubits)):
            cl_reg.append(compiled_qc.add_clbit())

        cl_reg_list.append(cl_reg)

        # Add measurement instruction
        compiled_qc.measure(qubits, cl_reg)

    # counts = execute(qs_temp, backend, basis_gates = basis_gates,
    # noise_model = noise_model, shots = shots).result().get_counts()
    counts = backend.run(compiled_qc, shots)
    counts = {k: counts[k] for k in sorted(counts)}

    # Convert the labeling bistrings of counts into list of labels
    new_counts = {}
    for i in range(len(counts)):
        # Retrieve the separated strings of each measurement variable

        counts_strings = []
        counts_bitstring = list(counts.keys())[i]

        bitstring_adress = 0
        for j in range(len(cl_reg_list)):
            cl_reg = cl_reg_list[::-1][j]
            counts_strings.append(
                counts_bitstring[bitstring_adress : bitstring_adress + len(cl_reg)][
                    ::-1
                ]
            )
            bitstring_adress += len(cl_reg)

        # Convert to integers and insert outcome labels
        counts_values = []
        for j in range(len(counts_strings)):
            outcome_int = int(counts_strings[j][::-1], 2)
            try:
                label = qv_list[j].decoder(outcome_int)
                if isinstance(label, np.ndarray):
                    from qrisp import OutcomeArray

                    label = OutcomeArray(label)
                counts_values.append(label)
            except AttributeError:
                counts_values.append(outcome_int)

        # Create array
        array_state = tuple(counts_values)
        try:
            new_counts[array_state] = counts[list(counts.keys())[i]] / shots
        except TypeError:
            raise Exception(
                "Tried to create measurement outcome dic for QuantumVariable "
                "with unhashable labels"
            )
        # Append to the counts list
        # counts_list.append((array_state, counts[list(counts.keys())[i]]/shots))

    # Sort counts_list such the most probable values come first
    new_counts = dict(sorted(new_counts.items(), key=lambda item: -item[1]))

    return new_counts


# Function to apply a phase function of signature phase_function(x,y,z..) -> float
# which specifies the phase for each constellation of outcome labels of the quantum
# variables in qv_list.
def app_phase_function(qv_list, phase_function, t=1, **kwargs):
    # Prepare the list of index tuples
    # For this we first create a list of outcome indices of each qv first
    index_lists = [list(range(2**qv.size)) for qv in qv_list]

    # We now calculate the direct product in order to obtain every possible combination
    from itertools import product

    product_index_list = list(product(*index_lists))

    # The next step is to iterate over every combination in order to determine the
    # phases.
    phases = []
    for i in range(len(product_index_list)):
        # Calculate the outcome labels of the current constellation of indices
        labels = [
            qv_list[j].decoder(product_index_list[i][j]) for j in range(len(qv_list))
        ]

        # Calculate the phase
        phases.append(phase_function(*labels, **kwargs) * t)

    # Synthesize phase
    from qrisp.logic_synthesis import gray_phase_synth_qb_list

    gray_phase_synth_qb_list(
        qv_list[0].qs, sum([qv.reg[::-1] for qv in qv_list], []), phases
    )


def as_hamiltonian(hamiltonian):
    r"""
    Decorator that recieves a regular Python function (returning a float) and returns a
    function of QuantumVariables, applying phases based on the function's output.

    Parameters
    ----------
    hamiltonian : function
        A function of arbitrary (non-quantum) variables returning a float.

    Returns
    -------
    hamiltonian_application : function
        A function of QuantumVariables, which applies the phase dictated by the
        hamiltonian to the corresponding states.

    Examples
    --------


    In this example we will demonstrate how a phase function with multiple arguments can
    be synthesized. For this we will create a phase function which encodes the fourier
    transform of different integers on the QuantumFloat x conditioned on the value of a
    QuantumChar ch. We will then apply the inverse Fourier transform to x and measure
    the results.
    ::

        import numpy as np
        from qrisp import QuantumChar, QuantumFloat, QFT, h
        #Create Variables
        x_size = 3
        x = QuantumFloat(x_size, 0, signed = False)

        ch = QuantumChar()

        # Bring x into uniform superposition so the phase function application yields
        # a fourier transformed computation basis state
        h(x)


        #Bring ch into partial superposition (here |a> + |b> + |c> + |d>)
        h(ch[0])
        h(ch[1])

        from qrisp import multi_measurement, as_hamiltonian

        #In order to define the hamiltonian, we use regular Python syntax
        #The decorator "as_hamiltonian" turns it into a function 
        #that takes Quantum Variables as arguments. The decorator will add the 
        #keyword argument t to the function which mimics the t in exp(i*H*t)


        @as_hamiltonian
        def apply_multi_var_hamiltonian(ch, x):
            if ch == "a":
                k = 2
            elif ch == "b":
                k = 7
            elif ch == "c":
                k = 3
            else:
                k = 4

            #Return phase value
            #This is the phase distribution of the Fourier-transform
            #of the computational basis state |k>
            return k*x * 2*np.pi/2**x_size


        #Apply Hamiltonian
        apply_multi_var_hamiltonian(ch,x, t = 1)


        #Apply inverse Fourier transform
        QFT(x, inv = True)


    Acquire measurement results

    >>> multi_measurement([ch, x])
    {('a', 2): 0.25, ('b', 7): 0.25, ('c', 3): 0.25, ('d', 4): 0.25}

    We see that the measurement results correspond to what we specified in the
    Hamiltonian.

    In Bra-Ket notation, before applying the Hamiltonian, we are in the state

    .. math::

        \ket{\psi} = \frac{1}{\sqrt{4}}(\ket{a} + \ket{b} + \ket{c} + \ket{d})
        \left( \frac{1}{\sqrt{8}} \sum_{x = 0}^8 \ket{x} \right)

    We then apply the Hamiltonian:

    .. math::

        \text{exp}(i\text{H(ch, x)})\ket{\psi} = \frac{1}{\sqrt{32}}
        ( &\ket{a} \sum_{x = 0}^{2^3-1} \text{exp}(2x \frac{2 \pi i}{2^3}) \ket{x} \\
        +&\ket{b} \sum_{x = 0}^{2^3-1} \text{exp}(7x \frac{2 \pi i}{2^3}) \ket{x} \\
        +&\ket{c} \sum_{x = 0}^{2^3-1} \text{exp}(3x \frac{2 \pi i}{2^3}) \ket{x} \\
        +&\ket{d} \sum_{x = 0}^{2^3-1} \text{exp}(4x \frac{2 \pi i}{2^3}) \ket{x})

    For each branch, the QuantumFloat tensor-factor is in a Fourier-transformed
    computational basis state. Thus, if we apply the inverse QFT, we receive:

    .. math::

        &\text{QFT}^{-1}\text{exp}(i\text{H(ch, x)})\ket{\psi} \\
            = & \frac{1}{\sqrt{4}} (\ket{a} \ket{2} + \ket{b} \ket{7} +\ket{c} \ket{3}
            + \ket{d} \ket{4})

    """

    def hamiltonian_application(*args, t=1, **kwargs):
        gate_wrap(app_phase_function)(args, hamiltonian, t=t, **kwargs)
        # app_phase_function(args, hamiltonian, t = t, **kwargs)

    return hamiltonian_application


def perm_lock(qubits, message=""):
    """
    Locks a list of qubits such that only permeable gates can be executed on these
    qubits. This means that an error will be raised if the user attempts to perform any
    operation involving these qubits if the operation does not commute with the
    Z-operator of this qubit. For more information, what a permeable gate is, check the
    :ref:`uncomputation documentation <uncomputation>`.

    This can be helpfull as it forbids all operations that change that computational
    basis state of this qubit but still allow controling on this qubit or applying
    phase gates.

    Using the keywoard message it is possible to extend the displayed error message.

    The effect of this function can be reversed using perm_unlock.

    Parameters
    ----------
    qubits : list[Qubit] or QuantumVariable
        The qubits to phase-tolerantly lock.
    message : str, optional
        The message why these qubits are locked.

    Examples
    --------

    We create a QuantumChar, perm-lock it's Qubits and attempt to initialize.

    >>> from qrisp import QuantumChar, perm_lock, cx, p
    >>> q_ch_0 = QuantumChar()
    >>> perm_lock(q_ch_0)
    >>> q_ch_0[:] = "g"
    Exception: Tried to perform non-permeable operations on perm_locked qubits

    We now create a second QuantumChar and perform a CNOT gate

    >>> q_ch_1 = QuantumChar()
    >>> cx(q_ch_0[3], q_ch_1[2])

    Phase-gates are possible, too

    >>> p(0.1, q_ch_0)

    """
    from qrisp.circuit.quantum_circuit import convert_to_qb_list

    for qb in convert_to_qb_list(qubits):
        if isinstance(qb, list):
            for item in qb:
                perm_lock(item)

            continue
        qb.perm_lock = True
        qb.perm_lock_message = message


def perm_unlock(qubits):
    """
    Reverses the effect of "perm_lock".

    Parameters
    ----------
    qubits : list[Qubit] or QuantumVariable
        The qubits to phase-tolerantly unlock.

    Examples
    --------

    We create a QuantumChar, perm-lock it's Qubits and attempt to initialize.

    >>> from qrisp import QuantumChar, perm_lock, perm_unlock
    >>> q_ch = QuantumChar()
    >>> perm_lock(q_ch, message = "Qubits are perm-locked due to testing purposes")
    >>> q_ch[:] = "g"
    Exception: Qubits are perm-locked due to testing purposes

    >>> perm_unlock(q_ch)
    >>> q_ch[:] = "g"
    >>> print(q_ch)
    {'g': 1.0}

    """
    from qrisp.circuit.quantum_circuit import convert_to_qb_list

    for qb in convert_to_qb_list(qubits):
        if isinstance(qb, list):
            for item in qb:
                perm_unlock(item)
            continue
        qb.perm_lock = False
        if hasattr(qb, "perm_lock_message"):
            del qb.perm_lock_message


def lock(qubits, message=""):
    """
    Locks a list of qubits, implying an error will be raised if the user tries to
    perform any operation involving these qubits.
    Using the keywoard message it is possible to extend the displayed error message.

    This can be reversed by calling unlock.

    Parameters
    ----------
    qubits : list[Qubit] or QuantumVariable
        The list of Qubits to lock.
    message : str, optional
        The message why these qubits are locked.

    Examples
    --------

    We create a QuantumChar, lock it's Qubits and attempt to initialize.

    >>> from qrisp import QuantumChar, lock
    >>> q_ch = QuantumChar()
    >>> lock(q_ch, message = "Qubits are locked due to testing purposes")
    >>> q_ch[:] = "g"
    Exception: Qubits are locked due to testing purposes

    """
    from qrisp.circuit.quantum_circuit import convert_to_qb_list

    for qb in convert_to_qb_list(qubits):
        if isinstance(qb, list):
            for item in qb:
                lock(item)
            continue

        qb.lock = True
        qb.lock_message = message


def unlock(qubits):
    """
    Reverses the effect of "lock".

    Parameters
    ----------
    qubits : list[Qubit] or QuantumVariable
        The list of Qubits to lock.

    Examples
    --------

    We create a QuantumChar, lock it's Qubits and attempt to initialize.

    >>> from qrisp import QuantumChar, lock, unlock
    >>> q_ch = QuantumChar()
    >>> lock(q_ch)
    >>> q_ch[:] = "g"
    Exception: Tried to perform operations on locked qubits

    We now unlock and try again

    >>> unlock(q_ch)
    >>> q_ch[:] = "g"
    >>> print(q_ch)
    {'g': 1.0}

    """
    from qrisp.circuit.quantum_circuit import convert_to_qb_list

    for qb in convert_to_qb_list(qubits):
        if isinstance(qb, list):
            for item in qb:
                unlock(item)
            continue
        qb.lock = False
        if hasattr(qb, "lock_message"):
            del qb.lock_message


def benchmark_function(function):
    def benchmarked_function(*args, sort_stats="tottime", stat_amount=20, **kwargs):
        def slow_function():
            function(*args, **kwargs)

        import cProfile
        import pstats

        profile = cProfile.Profile()

        profile.runcall(slow_function)

        ps = pstats.Stats(profile)
        ps.strip_dirs()
        ps.sort_stats(sort_stats)
        ps.print_stats(stat_amount)

    return benchmarked_function


def custom_qv(labels, decoder=None, qs=None, name=None):
    if not isinstance(labels, list):
        raise Exception(
            "Tried to create custom QuantumVariable without providing a list type"
        )

    if len(labels) == 0:
        raise Exception(
            "Tried to create custom QuantumVariable without providing labels"
        )
    elif len(labels) == 1:
        n = 1
    else:
        n = int(np.ceil(np.log2(len(labels))))

    from qrisp import QuantumVariable

    class CustomQuantumVariable(QuantumVariable):
        def __init__(self, qs=None, name=None):
            super().__init__(n, qs=qs, name=name)

        def decoder(self, x):
            if decoder is None:
                if x < len(labels):
                    return labels[x]
                else:
                    return "undefined_label_" + str(x)

            return decoder(x)

    return CustomQuantumVariable(qs=qs, name=name)


def init_state(qv, target_array):
    from qiskit.circuit.library.data_preparation.state_preparation import (
        StatePreparation,
    )

    qiskit_qc = StatePreparation(target_array).definition
    from qrisp import QuantumCircuit

    init_qc = QuantumCircuit.from_qiskit(qiskit_qc)

    # Find global phase correction
    from qrisp.simulator import statevector_sim

    init_qc.qubits.reverse()
    sim_array = statevector_sim(init_qc)
    init_qc.qubits.reverse()

    arg_max = np.argmax(np.abs(sim_array))

    gphase_dif = (np.angle(target_array[arg_max] / sim_array[arg_max])) % (2 * np.pi)

    init_qc.gphase(gphase_dif, 0)

    init_gate = init_qc.to_gate()

    init_gate.name = "state_init"

    qv.qs.append(init_gate, qv)


def get_statevector_function(qs, decimals=None):
    if len(qs.qv_list) == 0:
        return lambda x: 0
    else:
        from qrisp.simulator import statevector_sim

        compiled_qc = qs.compile()
        sv_array = statevector_sim(compiled_qc)

        if decimals is not None:
            sv_array = np.round(sv_array, decimals)

        def statevector(label_constellation, round=None):
            from qrisp import bin_rep

            qs = list(label_constellation.keys())[0].qs

            if len(label_constellation) != len(qs.qv_list):
                missing_variables = set([qv.name for qv in qs.qv_list]) - set(
                    [qv.name for qv in label_constellation.keys()]
                )
                raise Exception(
                    "Tried to invoke statevector debugger without specifying an "
                    "outcome label for each QuantumVariable registered in "
                    "QuantumSession. Missing variables are: "
                    + str(missing_variables)
                )

            bitstring = len(compiled_qc.qubits) * ["0"]

            for qf in label_constellation.keys():
                label_int = qf.encoder(label_constellation[qf])
                bin_label_int = bin_rep(label_int, qf.size)[::-1]

                for i in range(qf.size):
                    qubit_pos = compiled_qc.qubits.index(qf[i])
                    bitstring[qubit_pos] = bin_label_int[i]

            bitstring = "".join(bitstring)
            state_index = int(bitstring, base=2)

            if round is None:
                return sv_array[state_index]
            else:
                return np.around(sv_array[state_index], round)

        return statevector


def check_if_fresh(qubits, qs):
    qc = qs.transpile()

    reversed_data = qc.data[::-1]
    for qb in qubits:
        for instr in reversed_data:
            if qb in instr.qubits:
                if instr.op.name == "qb_alloc":
                    break
                else:
                    return False

    return True


def get_measurement_from_qc(qc, qubits, backend, shots=10000):
    # Add classical registers for the measurement results to be stored in
    cl = []
    for i in range(len(qubits)):
        cl.append(qc.add_clbit())

    # Add measurement instruction
    for i in range(len(qubits)):
        qc.measure(qubits[i], cl[i])

    # Execute circuit
    counts = backend.run(qc, shots)

    # Remove other measurements outcomes from counts dic
    new_counts_dic = {}
    for key in counts.keys():
        # Remove possible whitespaces
        new_key = key.replace(" ", "")
        # Remove other measurements
        new_key = new_key[-len(cl) :]

        new_key = int(new_key, base=2)
        try:
            new_counts_dic[new_key] += counts[key]
        except KeyError:
            new_counts_dic[new_key] = counts[key]

    counts = new_counts_dic
    # Plot result (if needed)

    # Normalize counts
    for key in counts.keys():
        counts[key] = counts[key] / abs(shots)

    return counts


def find_calling_line(level=0):
    stack = traceback.extract_stack(limit=level + 3)
    return str(
        traceback.format_list(stack)[1].split("\n")[1].strip()
    )  # prints "a = fct1()"


def retarget_instructions(data, source_qubits, target_qubits):
    from qrisp import QuantumEnvironment

    for i in range(len(data)):
        instr = data[i]

        if isinstance(instr, QuantumEnvironment):
            retarget_instructions(instr.original_data, source_qubits, target_qubits)
            retarget_instructions(instr.env_data, source_qubits, target_qubits)
            continue

        for j in range(len(instr.qubits)):
            if instr.qubits[j] in source_qubits:
                instr.qubits[j] = target_qubits[source_qubits.index(instr.qubits[j])]


def redirect_qfunction(function_to_redirect):
    """
    Decorator to turn a function returning a QuantumVariable into an in-place function.
    This can be helpful for manual uncomputation if we have a function returning some
    QuantumVariable, but we want the result to operate on some other variable, which is
    supposed to be uncomputed.

    Parameters
    ----------
    function_to_redirect : function
        A function returning a QuantumVariable.

    Raises
    ------
    Exception
        Given function did not return a QuantumVariable
    Exception
        Tried to redirect quantum function into QuantumVariable of differing size

    Returns
    -------
    redirected_function : function
        A function which performs the same operation as the input but now has the
        keyword argument target. Every instruction that would have been executed on the
        input functions result is executed on the QuantumVariable specified by target
        instead.


    Examples
    --------

    We create a function that determins the AND value of its inputs and redirect it
    onto another QuantumBool. ::

        from qrisp import QuantumBool, mcx, redirect_qfunction

        #This function has only two arguments and returns its result
        def AND(a, b):

            res = QuantumBool()

            mcx([a,b], res)

            return res

        a = QuantumBool(name = "a")
        b = QuantumBool(name = "b")
        c = QuantumBool(name = "c")

        #This function has two arguments and the keyword argument target
        redirected_AND = redirect_qfunction(AND)

        redirected_AND(a, b, target = c)


    >>> print(a.qs)
    QuantumCircuit:
    ---------------
    b.0: ──■──
           │
    a.0: ──■──
         ┌─┴─┐
    c.0: ┤ X ├
         └───┘
    Live QuantumVariables:
    ----------------------
    QuantumBool b
    QuantumBool a
    QuantumBool c



    """
    from qrisp import QuantumEnvironment, QuantumVariable

    def redirected_qfunction(*args, target=None, **kwargs):
        env = QuantumEnvironment()
        env.manual_allocation_management = True

        with env:
            res = function_to_redirect(*args, **kwargs)

            if not isinstance(res, QuantumVariable):
                raise Exception("Given function did not return a QuantumVariable")

            target = list(target)

            if len(res) != len(target):
                raise Exception(
                    "Tried to redirect quantum function into QuantumVariable of "
                    "differing size"
                )

            i = 0
            res_is_new = False
            while i < len(env.env_qs.data):
                instr = env.env_qs.data[i]
                if isinstance(instr, QuantumEnvironment):
                    pass
                elif instr.op.name == "qb_alloc" and instr.qubits[0] in list(res):
                    env.env_qs.data.pop(i)
                    res_is_new = True
                    continue

                i += 1

            retarget_instructions(env.env_qs.data, list(res), target)

        if res_is_new:
            # Remove all traces of res
            res.delete()

            for i in range(res.size):
                res.qs.qubits.remove(res[i])
                res.qs.data.pop(-1)

            for i in range(len(res.qs.deleted_qv_list)):
                qv = res.qs.deleted_qv_list[i]
                if qv.name == res.name:
                    res.qs.deleted_qv_list.pop(i)
                    break

        return target

    redirected_qfunction.__name__ = function_to_redirect.__name__

    return redirected_qfunction


def get_sympy_state(qs, decimals):
    from sympy import (
        I,
        Rational,
        Symbol,
        cancel,
        cos,
        count_ops,
        exp,
        factor,
        nsimplify,
        pi,
        simplify,
        sin,
    )
    from sympy.physics.quantum import Ket, OrthogonalKet

    from qrisp.simulator import statevector_sim

    qv_list = list(qs.qv_list)
    
    labels = []
    for qv in qv_list:
        labels.append([qv.decoder(i) for i in range(2**qv.size)])

    compiled_qc = qs.compile()

    sv_array = statevector_sim(compiled_qc)

    if not sv_array.dtype == np.dtype("O"):
        angles = np.angle(sv_array) % (2 * np.pi) / (np.pi)

        if decimals is not None:
            sv_array = np.round(sv_array, decimals)
            angles = np.round(angles, decimals)

        nz_indices = np.nonzero(sv_array)[0]
        nnz = len(nz_indices)

    else:
        import sympy as sp

        nz_indices = []

        for i in range(len(sv_array)):
            entry = simplify(sv_array[i])

            for a in sp.preorder_traversal(entry):
                if isinstance(a, sp.Float):
                    entry = entry.subs(a, round(a, decimals))

            sv_array[i] = entry

            if not sv_array[i] == 0:
                nz_indices.append(i)

        nnz = len(nz_indices)

    res = 0
    for ind in list(nz_indices):
        amplitude = sv_array[ind]

        if not sv_array.dtype == np.dtype("O"):
            abs_amp = trigify_amp(amplitude, decimals, nnz)

            # For some reason there is a sympy error, when the angle is equal to 1
            if angles[ind] == 1:
                phase = 1
            else:
                phase = nsimplify(angles[ind], tolerance=10 ** (decimals - 1))

            if count_ops(phase) > 5:
                phase = angles[ind]

            ket_expr = exp(I * phase * pi) * abs_amp * nnz**0.5

        else:
            process_stack = [amplitude]
            while process_stack:
                a = process_stack.pop(0)
                if (
                    isinstance(a, (sympy.core.add.Add, sympy.core.mul.Mul))
                    and len(a.free_symbols) != 0
                ):
                    process_stack.extend(a.args)

                elif len(a.free_symbols) == 0:
                    sub_float = np.round(complex(a.evalf()), decimals)

                    if np.abs(sub_float - 1) < 10**-decimals:
                        abs_amp = 1
                        continue
                    elif np.abs(sub_float) < 10**-decimals:
                        entry = entry.subs(a, 0)
                        continue
                    elif np.abs(sub_float) > 1:
                        continue
                    else:
                        abs_amp = trigify_amp(sub_float, decimals, nnz)

                    if np.angle(complex(a.evalf())) / np.pi == 1:
                        phase = -1
                    else:
                        phase = sp.exp(
                            sp.I
                            * nsimplify(
                                np.angle(complex(a.evalf())) / np.pi,
                                tolerance=10 ** (decimals - 1),
                            )
                            * Symbol("pi")
                        )

                    expr = abs_amp * phase

                    amplitude = amplitude.subs(a, expr)

            amplitude = amplitude.subs(1j, sp.I)

            ket_expr = sp.trigsimp(amplitude) * nnz**0.5

        int_string = bin_rep(ind, len(compiled_qc.qubits))

        labels = []
        for qv in qv_list:
            bit_string = ""
            for qb in qv.reg:
                bit_string += int_string[compiled_qc.qubits.index(qb)]

            label = qv.decoder(int(bit_string[::-1], 2))
            ket_expr *= OrthogonalKet((label))

        res += ket_expr

    res = cancel(nsimplify(1 / nnz**0.5) * res)

    if isinstance(res, sympy.core.mul.Mul):
        temp = 1
        for arg in res.args[:-1]:
            temp *= nsimplify(arg.subs({Symbol("pi"): pi}))

        res = temp * res.args[-1]

    res = res.subs({Symbol("pi"): pi})
    return res


def trigify_amp(amplitude, decimals, nnz):
    from sympy import (
        I,
        Rational,
        Symbol,
        cancel,
        cos,
        count_ops,
        exp,
        factor,
        latex,
        nsimplify,
        pi,
        simplify,
        sin,
    )

    cos_expr = nsimplify(
        np.arccos(np.abs(amplitude)) / np.pi, tolerance=10 ** -(decimals - 1)
    )
    sin_expr = nsimplify(
        np.arcsin(np.abs(amplitude)) / np.pi, tolerance=10 ** -(decimals - 1)
    )

    # if count_ops(sin_expr) > count_ops(cos_expr):
    if len(latex(sin_expr)) > len(latex(cos_expr)):
        expr = "cos"
        temp = cos_expr
    # elif count_ops(sin_expr) < count_ops(cos_expr):
    elif len(latex(sin_expr)) < len(latex(cos_expr)):
        expr = "sin"
        temp = sin_expr
    elif len(sin_expr.free_symbols) == 0:
        if sin_expr.evalf() > cos_expr.evalf():
            expr = "cos"
            temp = cos_expr
        else:
            expr = "sin"
            temp = sin_expr
    else:
        temp = (
            nsimplify(np.abs(amplitude) * nnz**0.5, tolerance=10 ** -(decimals - 1))
            / nnz**0.5
        )

    # if count_ops(temp) > 4:
    if len(latex(temp)) > 20:
        temp = (
            nsimplify(np.abs(amplitude) * nnz**0.5, tolerance=10 ** -(decimals - 1))
            / nnz**0.5
        )
        if len(latex(temp)) > 20:
            abs = np.abs(amplitude)

        else:
            abs = temp

    else:
        if expr == "cos":
            abs = cos(cos_expr * Symbol("pi"))
        else:
            abs = sin(sin_expr * Symbol("pi"))

    return abs


def render_qc(qc):
    latex_str = qc.to_latex()
    import os.path
    import subprocess
    import tempfile

    from IPython.display import Image, display

    with tempfile.TemporaryDirectory(prefix="texinpy_") as tmpdir:
        path = os.path.join(tmpdir, "document.tex")
        with open(path, "w") as fp:
            fp.write(latex_str)
        subprocess.run(["lualatex", path], cwd=tmpdir)
        subprocess.run(
            [
                "pdftocairo",
                "-singlefile",
                "-transp",
                "-r",
                "100",
                "-png",
                "document.pdf",
                "document",
            ],
            cwd=tmpdir,
        )

        im = Image(filename=os.path.join(tmpdir, "document.png"))
        display(im)


def lifted(*args, verify=False):
    """
    Shorthand for ``gate_wrap(permability = "args", is_qfree = True)``.

    A lifted function is ``qfree`` and permeable on its inputs. The results of lifted
    functions can be automatically uncomputed even if they contain functions that could
    not be uncomputed on their own.

    You can find more information about these concepts :ref:`here <Uncomputation>` or
    `here <https://silq.ethz.ch/overview#/overview/3_uncomputation>`_. Note that the
    concept of permeability in Qrisp is a more general version of Silq's ``const``.

    .. warning::

        Incorrect information about permeability and ``qfree``-ness can yield incorrect
        compilation results. If you are unsure, use the ``verify`` keyword on a small
        scale first.


    Parameters
    ----------

    verify : bool, optional
        If set to ``True``, the specified information about permeability and
        ``qfree``-ness will be checked numerically. The default is ``False``.

    Examples
    --------

    We create a function performing the `Margolus gate
    <https://arxiv.org/abs/quant-ph/0312225>`_. As it contains ``ry`` rotations,
    there are non-``qfree`` steps involved. Putting on the ``lifted`` decorator however
    marks the function as ``qfree`` as a whole.

    ::

        from qrisp import QuantumVariable, cx, ry, lifted
        from numpy import pi

        @lifted(verify = True)
        def margolus(control):

            res = QuantumVariable(1)
            ry(pi/4, res)
            cx(control[1], res)
            ry(-pi/4, res)
            cx(control[0], res)
            ry(pi/4, res)
            cx(control[1], res)
            ry(-pi/4, res)

            return res


        control = QuantumVariable(2)
        res = margolus(control)

    >>> print(res.qs)
    QuantumCircuit:
    ---------------
               ┌───────────┐
    control.0: ┤0          ├
               │           │
    control.1: ┤1 margolus ├
               │           │
        res.0: ┤2          ├
               └───────────┘
    Live QuantumVariables:
    ----------------------
    QuantumVariable control
    QuantumVariable res

    >>> res.uncompute()
    >>> print(res.qs)
    QuantumCircuit:
    ---------------
               ┌───────────┐┌──────────────┐
    control.0: ┤0          ├┤0             ├
               │           ││              │
    control.1: ┤1 margolus ├┤1 margolus_dg ├
               │           ││              │
        res.0: ┤2          ├┤2             ├
               └───────────┘└──────────────┘
    Live QuantumVariables:
    ----------------------
    QuantumVariable control

    Note that we set the ``verify`` keyword to ``True`` in this example. In more complex
    functions, involving many qubits this feature should only be used for bug-fixing on
    a small scale, since the verification can be time-consuming.

    """

    if len(args) == 0:

        def lifted_helper(function):
            return gate_wrap(permeability="args", is_qfree=True, verify=verify)(
                function
            )

        return lifted_helper

    else:
        return gate_wrap(permeability="args", is_qfree=True)(args[0])



"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""

from qrisp import cyclic_shift, QuantumVariable, QuantumArray, QuantumFloat, OutcomeArray
import numpy as np
def test_cycling_function():
    
    n = 3
    
    
    for l in range(5):
        for k in range(3):
            
            qa = QuantumArray(QuantumVariable(1), 2**n)
            qa[0][:] = "1"
            
            for i in range(l):
                cyclic_shift(qa, k)
    
            assert qa[(k*l)%(len(qa))].get_measurement() == {"1": 1}
            
            
            qa = QuantumArray(QuantumVariable(1), 2**n+k)
            qa[0][:] = "1"
            
            for i in range(l):
                cyclic_shift(qa)
    
            assert qa[(l)%(len(qa))].get_measurement() == {"1": 1}
            
    
    qa = QuantumArray(QuantumFloat(3), 8)

    qa[:] = np.arange(len(qa))

    shift_amount = QuantumFloat(2, signed = True, qs = qa.qs)

    shift_amount[:] = {0: 1, -4 : 1, 1: 1}
    cyclic_shift(qa, shift_amount)
    
    assert qa.get_measurement() == {OutcomeArray([0, 1, 2, 3, 4, 5, 6, 7]): 0.3333, OutcomeArray([7, 0, 1, 2, 3, 4, 5, 6]): 0.3333, OutcomeArray([4, 5, 6, 7, 0, 1, 2, 3]): 0.3333}



"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""


from qrisp.arithmetic import QuantumFloat
from qrisp.iterators.qrange import qRange
from qrisp import h, merge


def test_loop():
    from qrisp import QuantumFloat, qRange, h

    # Create some QuantumFloats
    n = QuantumFloat(3)
    qf = QuantumFloat(5)
    # Initialize the value 4
    n[:] = 4

    # Apply H-gate to 0-th qubit
    h(n[0])

    assert n.get_measurement() == {4.0: 0.5, 5.0: 0.5}

    # Perform successive addition of increasing numbers
    for i in qRange(n):
        qf += i

    assert qf.get_measurement() == {6.0: 0.5, 10.0: 0.5}



"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""

import setuptools

REQUIREMENTS = ["connexion>=2.12.0",
                "qiskit>=0.34.2",
                "thrift>=0.15.0",
                "matplotlib>=3.5.1",
                "waitress>=2.1.1",
                "scipy>=1.10.0",
                "numba",
                "networkx",
                "tdqm"]


with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="qrisp",
    author="Raphael Seidel",
    author_email="raphael.seidel@fokus.fraunhofer.de",
    description="Qrisp - A high level language for gate-based quantum computing",
    long_description=long_description,
    long_description_content_type="text/markdown",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Eclipse Public License 2.0 (EPL-2.0)",
        "Operating System :: OS Independent",
    ],
    package_dir={"": "src"},
    install_requires = REQUIREMENTS,
    setup_requires = REQUIREMENTS,
    packages=setuptools.find_packages(where="src"),
    python_requires=">=3.8",
)


"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""


import threading

from thrift.protocol import TBinaryProtocol
from thrift.server import TNonblockingServer, TServer
from thrift.transport import TSocket, TTransport

from qrisp import Qubit
from qrisp.interface import PortableQubit
from qrisp.interface.circuit_converter import convert_circuit
from qrisp.interface.thrift_interface.stoppable_thrift_server import (
    StoppableThriftServer,
)
from qrisp.quantum_network.interface.codegen.ttypes import Message
from qrisp.quantum_network.qn_simulator_server import QuantumNetwork
from qrisp.quantum_network.interface.codegen import QuantumNetworkService


# Returns the hosts ip
def get_ip():
    import socket

    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    s.settimeout(0)
    try:
        # doesn't even have to be reachable
        s.connect(("10.255.255.255", 1))
        IP = s.getsockname()[0]
    except Exception:
        IP = "127.0.0.1"
    finally:
        s.close()
    return IP


class QuantumNetworkHandler(QuantumNetwork):
    def run(self, qc, name):
        converted_circuit = convert_circuit(qc, "qrisp")
        return super().run(converted_circuit, name)

    def get_clear_qc(self, name):
        qc = super().get_clear_qc(name)

        return convert_circuit(qc, "thrift")

    def request_qubits(self, amount, name):
        qb_list = super().request_qubits(amount, name)

        return [PortableQubit(qb.identifier) for qb in qb_list]

    def send_qubits(self, sender, recipient, msg):
        super().send_qubits(
            sender,
            recipient,
            [Qubit(qb.identifier) for qb in msg.qubits],
            msg.annotation,
        )

    def get_overall_qc(self):
        qc = super().get_overall_qc()

        return convert_circuit(qc, "thrift")

    def inbox(self, name):
        inbox_list = super().inbox(name)

        res_list = []

        for i in range(len(inbox_list)):
            res_list.append(
                Message(
                    [PortableQubit(qb.identifier) for qb in inbox_list[i][0]],
                    inbox_list[i][1],
                )
            )

        return res_list


class QuantumNetworkServer:
    """
    This class sets up a server to coordinate the QuantumNetwork. All the simulations
    are performed in this instance, so the machine running it should preferably have
    abundant computational resources.

    Parameters
    ----------
    socket_ip : str, optional
        The IP address of the network socket. By default, the IP adress of the executing
        machine is used.
    port : int, optional
        The port to listen for requests. The default is 7070.
    background : bool, optional
        If set to True, the server will run in the background as a separate thread.
        The default is False.

    Examples
    --------

    We create a server listening to the localhost IP address:

    >>> from qrisp.quantum_network import QuantumNetworkServer
    >>> example_server = QuantumNetworkServer(socket_ip = "127.0.0.1")
    >>> example_server.start()

    """

    def __init__(self, socket_ip=None, port=None, background=False):
        if socket_ip is None:
            socket_ip = get_ip()

        self.socket_ip = socket_ip

        self.background = background
        self.port = port

    # Starts the server
    def start(self):
        """
        Starts the server.

        Returns
        -------
        None.

        """

        if self.port is None:
            self.port = 7070

        self.thread_name = "qn" + ":" + str(self.port)

        self.handler = QuantumNetworkHandler()

        processor = QuantumNetworkService.Processor(self.handler)
        tfactory = TTransport.TBufferedTransportFactory()

        self.transport = TSocket.TServerSocket(host=self.socket_ip, port=self.port)

        pfactory = TBinaryProtocol.TBinaryProtocolFactory()

        # server = TServer.TSimpleServer(processor, transport, tfactory, pfactory)
        # self.server = TServer.TThreadedServer(processor, self.transport, tfactory,
        # pfactory, daemon = True)
        self.server = StoppableThriftServer(
            processor, self.transport, tfactory, pfactory, daemon=True
        )

        # self.server = TNonblockingServer.TNonblockingServer(processor, self.transport,
        # tfactory, pfactory)

        if self.background:
            # Create thread
            self.thr = threading.Thread(target=self.server.serve, name=self.thread_name)
            self.thr.setDaemon(True)

            # Start the thread
            self.thr.start()
        else:
            self.server.serve()

    def stop(self):
        """
        Stops the server.

        Returns
        -------
        None.

        """
        if hasattr(self, "server"):
            self.server.stop()

    def __del__(self):
        self.handler.reset_network_state()
        self.stop()



"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""


import numpy as np
import sympy as sp

from qrisp import QuantumArray, p, z
from qrisp.arithmetic.SBP_arithmetic import hybrid_mult, polynomial_encoder
from qrisp.qtypes import QuantumFloat


def q_matmul(
    q_array_0, q_array_1, output_array=None, res_bit_shape="eq", phase_tolerant=False
):
    """
    Matrix multiplication for QuantumArrays.

    Parameters
    ----------
    q_array_0 : QuantumArray
        The first factor of the matrix multiplication.
    q_array_1 : QuantumArray
        The second factor of the matrix multiplication.
    output_array : QuantumArray, optional
        The QuantumArray to store the results in.
        By default, a new QuanumArray is created.
    res_bit_shape : str or QuantumFloat, optional
        Specification of the dimension of the output bitshape
        of the output QuantumArray.
        Possible are "eq", which will take the bitshape equal two the first factor,
        "safe" which automatically determines the bitshape such that there can be
        no overflow, or a QuantumFloat which has the desired bitshape.
        The default is "eq".
    phase_tolerant : bool, optional
        If set to True, the required gate count is reduced but each constellation of
        computational basis states of the inputs will introduce a different phase.
        This is helpful when it's clear that this function will be at some point
        uncomputed, resulting in the cancelation of these phases. The default is False.

    Raises
    ------
    Exception
        Tried to perform matrix multiplication with differing contraction index size.

    Returns
    -------
    res : QuantumArray
        The result of the matrix multiplication.

    Examples
    --------

    We multiply a QuantumArray with a multiply of the identity matrix (np.eye):

    >>> import numpy as np
    >>> from qrisp import QuantumFloat, QuantumArray, q_matmul
    >>> qf = QuantumFloat(4,0, signed = False)
    >>> q_arr_0 = QuantumArray(qf, shape = (2,2))
    >>> q_arr_1 = QuantumArray(qf, shape = (2,2))
    >>> q_arr_0[:] = [[1,2],[3,4]]
    >>> q_arr_1[:] = 2*np.eye(2)
    >>> res = q_matmul(q_arr_0, q_arr_1)
    >>> print(res)
    {OutcomeArray([[2, 4],
                   [6, 8]]): 1.0}

    """

    if q_array_0.shape[1] != q_array_1.shape[0]:
        raise Exception(
            "Tried to perform matrix multiplication"
            "with differing contraction index size"
        )

    L = q_array_0.shape[0]
    K = q_array_0.shape[1]
    J = q_array_1.shape[1]

    if output_array is not None:
        res = output_array

    else:
        from qrisp import QuantumFloat

        if isinstance(res_bit_shape, QuantumFloat):
            qtype = res_bit_shape.duplicate()

        if res_bit_shape == "safe":
            from sympy import Symbol

            q_array_0_symbols = []
            q_array_1_symbols = []
            poly = 0
            for i in range(K):
                q_array_0_symbols.append(Symbol(q_array_0[0, i].name))
                q_array_1_symbols.append(Symbol(q_array_1[i, 0].name))

                poly += Symbol(q_array_0[0, i].name) * Symbol(q_array_1[i, 0].name)
                # poly += q_array_0_symbols[-1]*q_array_1_symbols[-1]

            from qrisp.arithmetic import create_output_qf

            qtype = create_output_qf(
                list(q_array_0.flatten()) + list(q_array_1.flatten()), poly
            )

        if res_bit_shape == "eq":
            qtype = q_array_0[0, 0]

        res = QuantumArray(shape=(L, J), qtype=qtype, qs=q_array_0.qs)

    for i in range(L):
        for j in range(J):
            for k in range(K):
                if K == 1:
                    if output_array is not None:
                        hybrid_mult(
                            q_array_0[i, k],
                            q_array_1[k, j],
                            output_qf=res[i, j],
                            init_op="qft",
                            terminal_op="qft",
                            phase_tolerant=phase_tolerant,
                        )
                    else:
                        hybrid_mult(
                            q_array_0[i, k],
                            q_array_1[k, j],
                            output_qf=res[i, j],
                            init_op="h",
                            terminal_op="qft",
                            phase_tolerant=phase_tolerant,
                        )
                    continue

                # If the res_bit_shape is safe,we can use the strategy
                # of correcting phases on the result as displayed in hybrid mult
                if res_bit_shape == "safe" and not phase_tolerant:
                    if k == 0:
                        hybrid_mult(
                            q_array_0[i, k],
                            q_array_1[k, j],
                            output_qf=res[i, j],
                            init_op="h",
                            terminal_op=None,
                            phase_tolerant=True,
                        )

                    elif k == K - 1:
                        hybrid_mult(
                            q_array_0[i, k],
                            q_array_1[k, j],
                            output_qf=res[i, j],
                            init_op=None,
                            terminal_op="qft",
                            phase_tolerant=True,
                        )

                    elif k != q_array_0.shape[1] - 1:
                        hybrid_mult(
                            q_array_0[i, k],
                            q_array_1[k, j],
                            output_qf=res[i, j],
                            init_op=None,
                            terminal_op=None,
                            phase_tolerant=True,
                        )

                        for m in range(res[i, j].size):
                            p(
                                -(2 * np.pi) * 2**l / 2 ** (res[i, j].size + 1),
                                res[i, j][m],
                            )

                        if res[i, j].signed and K % 2:
                            z(res[i, j][-1])

                else:
                    if k == 0:
                        if output_array is not None:
                            hybrid_mult(
                                q_array_0[i, k],
                                q_array_1[k, j],
                                output_qf=res[i, j],
                                init_op="qft",
                                terminal_op=None,
                                phase_tolerant=phase_tolerant,
                            )
                        else:
                            hybrid_mult(
                                q_array_0[i, k],
                                q_array_1[k, j],
                                output_qf=res[i, j],
                                init_op="h",
                                terminal_op=None,
                                phase_tolerant=phase_tolerant,
                            )
                    elif k == K - 1:
                        hybrid_mult(
                            q_array_0[i, k],
                            q_array_1[k, j],
                            output_qf=res[i, j],
                            init_op=None,
                            terminal_op="qft",
                            phase_tolerant=phase_tolerant,
                        )
                    else:
                        hybrid_mult(
                            q_array_0[i, k],
                            q_array_1[k, j],
                            output_qf=res[i, j],
                            init_op=None,
                            terminal_op=None,
                            phase_tolerant=phase_tolerant,
                        )

    return res


def semi_classic_matmul(q_matrix, cl_matrix, output_array=None, res_bit_shape="eq"):
    """
    Performs matrix multiplication between a classical numpy array and a QuantumArray

    Parameters
    ----------
    q_matrix : QuantumArray
        The QuantumArray to multiply.
    cl_matrix : numpy.ndarray
        The numpy array to multiply.
    output_array : QuantumArray, optional
        The QuantumArray to store the result in. The default is None.
    res_bit_shape : str or QuantumFloat, optional
        Specification of the dimension of the output bitshape
        of the output QuantumArray.
        Possible are "eq", which will take the bitshape equal two the first factor,
        "safe" which automatically determines the bitshape such that there can be no
        overflow, or a QuantumFloat which has the desired bitshape. The default is "eq".

    Raises
    ------
    Exception
        Tried to apply matrix multiplication with unfitting dimensions.

    Returns
    -------
    output_array : QuantumArray
        The QuantumArray containing the result.

    Examples
    --------

    We multiply a QuantumArray with a scalar multiple of the identity matrix (np.eye):

    >>> import numpy as np
    >>> from qrisp import QuantumFloat, QuantumArray, semi_classic_matmul
    >>> qf = QuantumFloat(4,0, signed = False)
    >>> q_arr_0 = QuantumArray(qf, shape = (2,2))
    >>> q_arr_1 = QuantumArray(qf, shape = (2,2))
    >>> q_arr_0[:] = [[1,2],[3,4]]
    >>> res = semi_classic_matmul(q_arr_0, 2*np.eye(2))
    >>> print(res)
    {OutcomeArray([[2, 4],
                   [6, 8]]): 1.0}


    """
    if q_matrix.shape[1] != cl_matrix.shape[0]:
        raise Exception(
            "Tried to apply matrix multiplication with unfitting dimensions"
        )

    L = q_matrix.shape[0]
    K = q_matrix.shape[1]
    J = cl_matrix.shape[1]

    from sympy import Symbol

    if output_array is None:
        if isinstance(res_bit_shape, QuantumFloat):
            qtype = res_bit_shape.duplicate()

        if res_bit_shape == "safe":
            q_array_0_symbols = []
            poly = 0
            for i in range(K):
                q_array_0_symbols.append(Symbol(q_matrix[0, i].name))

                poly += Symbol(q_matrix[0, i].name)

            from qrisp.arithmetic import create_output_qf

            qtype = create_output_qf(list(q_matrix.flatten()), poly)

        if res_bit_shape == "eq":
            qtype = q_matrix[0, 0]

        output_array = QuantumArray(shape=(L, J), qtype=qtype, qs=q_matrix.qs)

    if output_array.shape != (L, J):
        raise Exception(
            "Tried to to encode matrix multiplication"
            "into Quantum Array of unfitting size"
        )

    from sympy.matrices import zeros

    symb_matrix = zeros(q_matrix.shape[0], q_matrix.shape[1])

    symbol_dic = {}
    for i in range(L):
        for k in range(K):
            temp_symb = Symbol("x_" + str(i) + "_" + str(k))
            symb_matrix[i, k] = temp_symb
            symbol_dic[q_matrix[i, k].name] = temp_symb

    res_symb_matrix = symb_matrix @ cl_matrix

    qv_list = list(q_matrix.flatten())
    for i in range(L):
        for j in range(J):
            if res_symb_matrix[i, j] != 0:
                polynomial_encoder(
                    qv_list,
                    output_array[i, j],
                    res_symb_matrix[i, j],
                    encoding_dic=symbol_dic,
                )

    return output_array


def egcd(a, b):
    if a == 0:
        return (b, 0, 1)
    else:
        g, y, x = egcd(b % a, a)
        return (g, x - (b // a) * y, y)


def modinv(a, m):
    g, x, y = egcd(a, m)
    if g != 1:
        raise Exception("modular inverse does not exist")
    else:
        return x % m


# Function for inplace multiplication with a quantum array
# [qv1, qv2,.. qvn] with a n x n matrix  A
# ie. the array contains [a11*qv1 +a12*qv2+ ... , a21*qv1..] after
def inplace_matrix_app(vector, matrix):
    r"""
    Performs inplace matrix application to a vector-valued QuantumArray.
    Note that due to reversibility reasons, the matrix can only contain integers
    and has to be invertible over $\text{GL}(2^n)$.
    This is equal to the condition that the determinant is odd.

    Parameters
    ----------
    vector : QuantumArray
        The QuantumArray to apply the matrix to.
    matrix : numpy.ndarray
        The matrix to apply.

    Raises
    ------
    Exception
        Tried to multiply matrix with Quantum Array with unfitting shape.

    Examples
    --------

    We perform an inplace matrix multiplication between a randomly chosen matrix
    and a randomly chosen vector

    >>> from qrisp import QuantumFloat, QuantumArray, inplace_matrix_app
    >>> qtype = QuantumFloat(4)
    >>> q_vector = QuantumArray(qtype, 3)
    >>> q_vector[:] = [1,0,2]
    >>> matrix = np.array([[2., 2., 1.],
                           [2., 3., 1.],
                           [3., 1., 2.]])
    >>> inplace_matrix_app(q_vector, matrix)
    >>> print(q_vector)
    {OutcomeArray([4, 4, 7]): 1.0}
    """

    if len(vector.shape) != 1:
        raise Exception(
            "Tried to multiply matrix with Quantum Array with unfitting shape"
        )

    n = vector.shape[0]

    if n != matrix.shape[0] or n != matrix.shape[1]:
        raise Exception(
            "Tried to multiply matrix with Quantum Array with unfitting shape"
        )

    bit = vector[0].size

    # The general idea is to calculate a row of the matrix into the variable x_j which
    # can be reconstructed from the result of the calculation and thus discarded.

    # Calling the result of this evaluation x_j_new we have

    # x_j_new = a*x_j + b*x_0 ... c*x_n

    # For this equation to be invertible a has to be invertible. We then have

    # x_j = 1/a*(x_j_new - b*x_0 + .. c*x_n)

    # we then replace x_j in the remaining evaluation equations with this expression

    # We will perform this process in two steps:
    # 1. Prepare a list of equations that contains what should be evaluated
    # 2. Evaluate these equations in the quantum session

    # Step 1:

    # Prepare symbols
    ancilla_symbol = sp.Symbol("a")
    x = [sp.Symbol("x" + str(i)) for i in range(n)]

    # This list contains the equations that need to be calculated
    target_values = matrix * sp.Matrix(x)

    # This list contains the variables that have been eliminated in favor
    # of the completed equations
    eliminated_variables = []

    # This list contains the equations
    # that will be evaluated later by the circuit generator
    eval_eq = []

    from qrisp.misc import is_inv

    for i in range(n):
        # Save evaluation
        eval_eq.append(target_values[i])

        # Find a column with invertible coeffiecient
        j = 0
        while True:
            # Determine that coefficient(can be done by differentiating the evaluating
            # eq. after the column variable)
            coeff = int(sp.diff(target_values[i], x[j]))

            # Check if the coefficient is invertible and has not been eliminated yet
            if is_inv(coeff, bit) and (j not in eliminated_variables):
                break
            j += 1

            if j >= n:
                raise Exception("Could not find invertible element")

        # Generate inverse equation
        eval_inverse = modinv(coeff, 2**bit) * (
            -target_values[i].subs({x[j]: 0}) + ancilla_symbol
        )

        from qrisp.qtypes.quantum_float import trunc_poly

        # Truncate coefficients in order to stay inside Z/ 2^n Z
        eval_inverse = trunc_poly(eval_inverse, (0, bit))

        # Rewrite the remaining evaluation equations in terms of the new variable
        subs_dic = {x[j]: eval_inverse}

        for k in range(i + 1, n):
            # Substitue in the following equations
            target_values[k] = (
                target_values[k].subs(subs_dic).subs({ancilla_symbol: x[j]})
            )

            # Truncate coefficients
            target_values[k] = trunc_poly(target_values[k], (0, bit))

        eliminated_variables.append(j)

    # Step 2:

    # Create symbol dic in order for the polynomial encoder
    # to know which symbol to use as which quantum float
    symbol_dic = {vector[j].name: x[j] for j in range(vector.shape[0])}

    for i in range(n):
        # Evaluate the value of the entry in eval_eq into the x_j variable

        elim_var = eliminated_variables[i]

        # Create list of quantum variables which are not being written on
        non_elim_var_list = list(vector)
        non_elim_var_list.pop(elim_var)

        # Perform evaluation. The first step is to determine the coefficient for the
        # inplace multiplication (ie. in the language of the initial description "a")
        inplace_mult_coeff = int(sp.diff(eval_eq[i], x[elim_var]))

        # Now determine the remaining polynomial
        eval_poly = eval_eq[i].subs({x[elim_var]: 0})
        print(inplace_mult_coeff)
        polynomial_encoder(
            non_elim_var_list,
            vector[elim_var],
            eval_poly,
            encoding_dic=symbol_dic,
            inplace_mult=inplace_mult_coeff,
        )

    # Reorder quantum array
    qv_reordering_array = np.zeros(n, dtype="object")

    for i in range(n):
        qv_reordering_array[i] = vector[eliminated_variables[i]]

    # In order to enter the reorder values in the QuantumArray, we need to use the
    # setitem method of ndarray because setitem for QuantumArray
    # is overloaded with initialization
    np.ndarray.__setitem__(vector, slice(None, None, None), qv_reordering_array)


def auto_matmul_wrapper(a, b, out=None):
    if isinstance(a, QuantumArray) and isinstance(b, QuantumArray):
        return q_matmul(a, b, out)

    elif isinstance(a, QuantumArray):
        return semi_classic_matmul(a, b, out)

    elif isinstance(b, QuantumArray):
        return semi_classic_matmul(b.transpose(), a.transpose(), out).transpose()

    else:
        raise Exception(
            "Could not proccess input constellation "
            + str(type(a))
            + " and "
            + str(type(b))
        )


def dot(a, b, out=None):
    """
    Port of the popular `numpy function
    <https://numpy.org/doc/stable/reference/generated/numpy.dot.html>`_
    with similar semantics.

    Parameters
    ----------
    a : QuantumArray or QuantumFloat or numpy.ndarray
        The first operand.
    b : QuantumArray or QuantumFloat or numpy.ndarray
        The second operand.
    out : QuantumArray, optional
        The QuantumArray to store the output in. The default is None.

    Returns
    -------
    QuantumArray
        The result as described in the numpy documentation.

    Examples
    --------

    We create two QuantumArrays and apply dot as a function
    performing matrix-vector multiplication.

    >>> import numpy as np
    >>> from qrisp import QuantumFloat, QuantumArray, dot
    >>> qf = QuantumFloat(5,0, signed = False)
    >>> q_arr_0 = QuantumArray(qf)
    >>> q_arr_1 = QuantumArray(qf)
    >>> q_arr_0[:] = [2,3]
    >>> q_arr_1[:] = 2*np.eye(2)
    >>> res = dot(q_arr_0, q_arr_1)
    >>> print(res)
    {OutcomeArray([[4, 6]]): 1.0}

    Scalar-product:

    >>> q_arr_0 = QuantumArray(qf)
    >>> q_arr_1 = QuantumArray(qf)
    >>> q_arr_0[:] = [3,4,5]
    >>> q_arr_1[:] = [1,1,1]
    >>> res = dot(q_arr_0, q_arr_1)
    >>> print(res)
    {12: 1.0}

    Matrix-matrix multiplication

    >>> qf = QuantumFloat(3,0, signed = True)
    >>> q_arr_0 = QuantumArray(qf)
    >>> q_arr_1 = QuantumArray(qf)
    >>> q_arr_0[:] = [[0,1],[1,0]]
    >>> q_arr_1[:] = [[1,0],[0,-1]]
    >>> res = dot(q_arr_0, q_arr_1)
    >>> print(res)
    {OutcomeArray([[ 0, -1],
                   [ 1,  0]]): 1.0}

    """
    from qrisp.arithmetic import QuantumFloat

    if isinstance(a, QuantumFloat) or isinstance(b, QuantumFloat):
        return np.dot(a, b, out)

    else:
        if len(a.shape) == 1 and len(b.shape) == 1:
            temp_0 = a.reshape((1, a.shape[0]))
            temp_1 = b.reshape((a.shape[0], 1))
            res = (auto_matmul_wrapper(temp_0, temp_1, out))[0, 0]
            return res

        elif len(a.shape) == 2 and len(b.shape) == 2:
            return auto_matmul_wrapper(a, b, out)

        elif len(b.shape) == 1:
            temp_0 = a.reshape((a.size // b.shape[-1], b.shape[-1]))
            temp_1 = b.reshape((b.shape[0], 1))

            new_shape = list(b.shape)
            new_shape[-1] = b.shape[0]

            res = (auto_matmul_wrapper(temp_0, temp_1, out)).reshape(new_shape)

            return res

        else:
            temp_0 = a.reshape((a.size // b.shape[-1], b.shape[-1]))

            temp_1 = np.swapaxes(b, 0, -2)
            temp_1 = np.reshape(
                temp_1, (temp_1.shape[0], temp_1.size // temp_1.shape[0])
            )

            res = auto_matmul_wrapper(temp_0, temp_1, out)

            res_shape = list(temp_1.shape)
            res_shape.pop(-2)

            res_shape = list(temp_0.shape)[:-1] + res_shape

            return np.reshape(res, res_shape)


def tensordot(a, b, axes):
    r"""
    Port of `numpy tensordot
    <https://numpy.org/doc/stable/reference/generated/numpy.tensordot.html>`_
    with similar semantics.

    Parameters
    ----------
    a : QuantumArray
        The first operand.
    b : QuantumArray
        The second operand.
    axes : tuple
        The axes to contract.

    Returns
    -------
    QuantumArray
        The QuantumArray containing the result of tensordot.

    Examples
    --------

    Using ``tensordot`` we can perform the arithmetic for simulating a quantum computer
    *on a quantum computer*.

    >>> import numpy as np
    >>> from qrisp import QuantumFloat, QuantumArray, tensordot

    Initiate the QuantumArray holding the statevector. We initate the state of uniform
    superposition

    .. math::

        \ket{+} = \frac{1}{\sqrt{2^n}} \sum_{i = 0}^{2^n - 1} \ket{i}

    >>> qfloat_type = QuantumFloat(3, -2, signed = True)
    >>> num_qubits = 4
    >>> statevector = QuantumArray(shape = 2**num_qubits, qtype = qfloat_type)
    >>> statevector[:] = [1/(2**num_qubits)**0.5]*2**num_qubits
    >>> print(statevector)
    {OutcomeArray([0.25, 0.25, 0.25, 0.25, 0.25, 0.25, 0.25, 0.25,
                   0.25, 0.25, 0.25, 0.25, 0.25, 0.25, 0.25, 0.25]): 1.0}

    Initiate the QuantumArray holding the unitary of a Z-gate

    >>> z_gate = QuantumArray(shape = (2,2), qtype = qfloat_type)
    >>> z_gate[:] = [[1,0], [0,-1]]
    >>> print(z_gate)
    {OutcomeArray([[ 1.,  0.],
                  [ 0., -1.]]): 1.0}

    Perform the contraction

    >>> statevector = statevector.reshape(num_qubits*[2])
    >>> target_qubit = 3
    >>> new_statevector = tensordot(z_gate, statevector, (1, target_qubit))
    >>> new_statevector = new_statevector.reshape(2**num_qubits)
    >>> print(new_statevector)
    {OutcomeArray([ 0.25,  0.25,  0.25,  0.25,  0.25,  0.25,  0.25,  0.25,
                   -0.25, -0.25, -0.25, -0.25, -0.25, -0.25, -0.25, -0.25]): 1.0}

    We perform a similar contraction using numpy arrays

    >>> from numpy import tensordot
    >>> statevector = 0.25*np.ones(2**num_qubits)
    >>> print(statevector)
    [0.25 0.25 0.25 0.25 0.25 0.25 0.25 0.25 0.25 0.25 0.25 0.25 0.25 0.25
     0.25 0.25]
    >>> statevector = statevector.reshape(num_qubits*[2])
    >>> z_gate = np.zeros((2,2))
    >>> z_gate[:] = [[1,0], [0,-1]]
    >>> new_statevector = tensordot(z_gate, statevector, (1, target_qubit))
    >>> print(new_statevector.reshape(2**num_qubits))
    [ 0.25  0.25  0.25  0.25  0.25  0.25  0.25  0.25 -0.25 -0.25 -0.25 -0.25
     -0.25 -0.25 -0.25 -0.25]

    """

    try:
        iter(axes)
    except Exception:
        axes_a = list(range(-axes, 0))
        axes_b = list(range(0, axes))
    else:
        axes_a, axes_b = axes
    try:
        na = len(axes_a)
        axes_a = list(axes_a)
    except TypeError:
        axes_a = [axes_a]
        na = 1
    try:
        nb = len(axes_b)
        axes_b = list(axes_b)
    except TypeError:
        axes_b = [axes_b]
        nb = 1

    a, b = np.asanyarray(a), np.asanyarray(b)
    as_ = a.shape
    nda = a.ndim
    bs = b.shape
    ndb = b.ndim
    equal = True

    if na != nb:
        equal = False
    else:
        for k in range(na):
            if as_[axes_a[k]] != bs[axes_b[k]]:
                equal = False
                break
            if axes_a[k] < 0:
                axes_a[k] += nda
            if axes_b[k] < 0:
                axes_b[k] += ndb
    if not equal:
        raise ValueError("shape-mismatch for sum")

    # Move the axes to sum over to the end of "a"
    # and to the front of "b"
    notin = [k for k in range(nda) if k not in axes_a]
    newaxes_a = notin + axes_a
    N2 = 1
    for axis in axes_a:
        N2 *= as_[axis]
    newshape_a = (int(np.multiply.reduce([as_[ax] for ax in notin])), N2)
    olda = [as_[axis] for axis in notin]

    notin = [k for k in range(ndb) if k not in axes_b]
    newaxes_b = axes_b + notin
    N2 = 1
    for axis in axes_b:
        N2 *= bs[axis]
    newshape_b = (N2, int(np.multiply.reduce([bs[ax] for ax in notin])))
    oldb = [bs[axis] for axis in notin]

    at = a.transpose(newaxes_a).reshape(newshape_a)
    bt = b.transpose(newaxes_b).reshape(newshape_b)

    res = dot(at, bt)
    return res.reshape(olda + oldb)



"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""

# Created by ann81984 at 22.07.2022
from qrisp import QuantumString, QuantumChar


def test_string_test():
    q_str = QuantumString()

    q_str_2 = QuantumString()

    q_ch = QuantumChar()

    q_str[:] = "hello"

    q_ch[:] = " "

    q_str_2[:] = "world"

    q_str += q_ch

    q_str += q_str_2

    q_str += "! "

    q_str_3 = q_str.duplicate(init=True)

    assert (type(q_str).__name__ and type(q_str_3).__name__) == "QuantumString"
    assert str(q_str + q_str_3) == "{'hello world! hello world! ': 1.0}"
    print(q_str + q_str_3)



"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""


def test_mcx():
    from qrisp import h, mcx, QuantumVariable, multi_measurement, bin_rep
    import numpy as np

    def test_mcx_inner(n, ctrl_state, method, test_phase=True, num_ancilla=0):
        ctrl = QuantumVariable(n)

        target = QuantumVariable(1)

        h(ctrl)

        if method in ["hybrid", "balauca_dirty"]:
            mcx(
                ctrl,
                target,
                method=method,
                ctrl_state=ctrl_state,
                num_ancilla=num_ancilla,
            )
        else:
            mcx(ctrl, target, method=method, ctrl_state=ctrl_state)
        mes_res = multi_measurement([ctrl, target])
        statevector = ctrl.qs.statevector("array")

        # Test correct flipping behavior
        for i in range(2**n):
            ctrl_var = bin_rep(i, n)
            if ctrl_var != ctrl_state:
                assert (ctrl_var, "0") in mes_res
            else:
                assert (ctrl_var, "1") in mes_res

        if not test_phase:
            return

        angles = np.angle(statevector[np.abs(statevector) > 1 / 2 ** (n / 2 + 1)])

        # Test correct phase behavior
        assert np.sum(np.abs(angles)) < 0.1

    for n in range(1, 8):
        print("n: ", n)
        for ctrl_state in [
            n * "1",
            n * "0",
            bin_rep(int(np.random.randint(0, 2**n - 1)), n),
        ]:
            print("Ctrl state: ", ctrl_state)
            for method in ["gray", "gray_pt", "maslov", "balauca", "yong"]:
                print(method)

                if method == "maslov" and n > 4:
                    continue

                test_mcx_inner(n, ctrl_state, method, test_phase=(method != "gray_pt"))

            for method in ["hybrid", "balauca_dirty"]:
                for num_ancilla in range(8):
                    print("Ancilla count: ", num_ancilla)
                    test_mcx_inner(n, ctrl_state, method, num_ancilla=num_ancilla)



"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""

# Created by ann81984 at 05.05.2022
from qrisp import QuantumVariable, x, h, merge
from qrisp.environments.GMS_environment import GMSEnvironment
import numpy as np


def test_GMS_environment_example():
    qv1 = QuantumVariable(1)
    qv2 = QuantumVariable(1)

    merge([qv1, qv2])

    qs = qv1.qs
    test = GMSEnvironment()

    h(qv1)
    x(qv2)
    with test:
        qs.cp(np.pi / 2, 0, 1)
        qs.p(np.pi / 2, 0)
        qs.p(np.pi / 2, 1)

    h(qv1)

    ###################
    from qrisp.arithmetic import QuantumFloat

    qf = QuantumFloat(3)

    qf.encode(1)

    from qrisp import QFT

    QFT(qf, use_gms=True)

    sv = qf.qs.statevector("array")



"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""


from qrisp.core import QuantumVariable


class QuantumChar(QuantumVariable):
    r"""
    A QuantumVariable which represents characters. By default, the QuantumChar is
    initialized in NISQ mode, meaning that instead of 256 characters it can only hold
    32, saving almost 40% in qubit cost.

    >>> from qrisp import QuantumChar
    >>> q_ch = QuantumChar(nisq_char = True)

    The chars which can be represented in ``nisq mode`` are

    +---+---+---+---+---+---+---+---+---+---+----+----+----+----+----+----+----+
    | 0 | 1 | 2 | 3 | 4 | 5 | 6 | 7 | 8 | 9 | 10 | 11 | 12 | 13 | 14 | 15 | 16 |
    +---+---+---+---+---+---+---+---+---+---+----+----+----+----+----+----+----+
    | a | b | c | d | e | f | g | h | i | j | k  | l  | m  | n  | o  | p  | q  |
    +---+---+---+---+---+---+---+---+---+---+----+----+----+----+----+----+----+

    +----+----+----+----+----+----+----+----+----+----+----+----+----+----+----+
    | 17 | 18 | 19 | 20 | 21 | 22 | 23 | 24 | 25 | 26 | 27 | 28 | 29 | 30 | 31 |
    +----+----+----+----+----+----+----+----+----+----+----+----+----+----+----+
    | r  | s  | t  | u  | v  | w  | x  | y  | z  |    | .  | !  | ?  | :  | ,  |
    +----+----+----+----+----+----+----+----+----+----+----+----+----+----+----+

    If ``nisq_mode`` is set to False, the encoder uses the Python-inbuild chr function.
    """

    def decoder(self, i):
        if self.nisq_char:
            return "abcdefghijklmnopqrstuvwxyz .!?:,"[i]
        else:
            return chr(i)

    def __init__(self, qs=None, name=None, nisq_char=True):
        self.nisq_char = nisq_char

        if nisq_char:
            super().__init__(5, qs=qs, name=name)
        else:
            super().__init__(8, qs=qs, name=name)



"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""


from qrisp import Clbit, QuantumCircuit, Qubit
from qrisp.simulator import ImpureQuantumState, QuantumState, run, single_shot_sim


class QuantumNetwork:
    def __init__(self):
        self.qs = QuantumState(0)
        self.qubit_map = {}
        self.overall_circuit = QuantumCircuit()
        self.client_measurement_counter = {}
        self.inbox_dict = {}
        self.qubit_creation_counter = {}

    def register_client(self, name):
        if name in list(self.client_measurement_counter.keys()):
            raise Exception("Tried to register an existing client.")

        self.client_measurement_counter[name] = 0
        self.inbox_dict[name] = []
        self.qubit_creation_counter[name] = 0

    def send_qubits(self, sender, recipient, qubits, annotation):
        keys = list(self.qubit_map.keys())
        ids = keys

        qubits = convert_qubit_names(qubits, sender)

        for qb in qubits:
            if qb.identifier in ids:
                index = ids.index(qb.identifier)

                if sender != self.qubit_map[keys[index]]:
                    raise Exception("Tried to send qubit not belonging to sender.")
            else:
                self.add_qubit(qb, sender)

            self.qubit_map[qb.identifier] = recipient

        self.inbox_dict[recipient].append((qubits, annotation))

    def get_clear_qc(self, name):
        qc = QuantumCircuit()

        for key in self.qubit_map.keys():
            if self.qubit_map[key] == name:
                qc.add_qubit(Qubit(key))

        return qc

    def request_qubits(self, amount, owner):
        res_list = []
        for i in range(amount):
            new_qb = Qubit(
                "qb_" + str(self.qubit_creation_counter[owner]) + "@" + owner
            )
            self.add_qubit(new_qb, owner)
            res_list.append(new_qb)

        return res_list

    def add_qubit(self, qubit, owner):
        if owner not in self.client_measurement_counter.keys():
            raise Exception("Tried to request qubits for non registered client")

        qubit = convert_qubit_names([qubit], owner)[0]

        if qubit in self.qubit_map.keys():
            raise Exception("Tried to create already existing qubit")

        self.qubit_map[qubit.identifier] = owner
        self.qubit_creation_counter[owner] += 1
        self.overall_circuit.add_qubit(qubit)
        self.qs.add_qubit()

        return [qubit]

    def run(self, qc, name):
        keys = list(self.qubit_map.keys())
        ids = keys

        qc = convert_qc_qubit_names(qc, name)

        for qb in qc.qubits:
            if qb.identifier in ids:
                index = ids.index(qb.identifier)

                if self.qubit_map[keys[index]] != name:
                    raise Exception("Tried to operate one foreign qubits")

            else:
                self.add_qubit(qb, name)

        temp_qc = self.overall_circuit.clearcopy()
        temp_qc.clbits = []

        translation_dic = {}

        for clbit in qc.clbits:
            overall_qc_clbit = Clbit(
                "cb" + "_" + name + "_" + str(self.client_measurement_counter[name])
            )
            self.client_measurement_counter[name] += 1

            self.overall_circuit.add_clbit(overall_qc_clbit)
            temp_qc.add_clbit(clbit)

            translation_dic[clbit] = overall_qc_clbit

        translation_dic.update({qb: qb for qb in qc.qubits})

        temp_qc.extend(qc)

        res, self.qs = single_shot_sim(temp_qc, self.qs)

        self.overall_circuit.barrier()
        self.overall_circuit.extend(qc, translation_dic)

        return {res: 1}

    def get_overall_qc(self):
        return self.overall_circuit.copy()

    def inbox(self, name):
        return self.inbox_dict[name]

    def reset_network_state(self):
        self.qs = QuantumState(0)
        self.qubit_map = {}
        self.overall_circuit = QuantumCircuit()
        self.client_measurement_counter = {}
        self.inbox = {}


def convert_qc_qubit_names(qc, name):
    res_qc = qc.clearcopy()

    res_qc.qubits = convert_qubit_names(qc.qubits, name)

    transl_dic = {
        qc.qubits[i].identifier: res_qc.qubits[i] for i in range(len(qc.qubits))
    }

    for instr in qc.data:
        res_qc.append(
            instr.op, [transl_dic[qb.identifier] for qb in instr.qubits], instr.clbits
        )

    return res_qc


def convert_qubit_names(qubits, name):
    res = []
    for i in range(len(qubits)):
        qb = qubits[i]
        if len(qb.identifier.split("@")) == 1:
            res.append(Qubit(qb.identifier + "@" + name))
        else:
            res.append(Qubit(qb.identifier))

    return res



"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""

import threading
import sys
# -*- coding: utf-8 -*-
import numpy as np
from tqdm import tqdm

from qrisp.circuit import (
    ControlledOperation,
    Instruction,
    Operation,
    QuantumCircuit,
    fast_append,
    transpile,
)
from qrisp.interface.circuit_converter import convert_circuit
from qrisp.simulator.circuit_preprocessing import (
    circuit_preprocessor,
    count_measurements_and_treat_alloc,
    group_qc,
)
from qrisp.simulator.impure_quantum_state import ImpureQuantumState
from qrisp.simulator.quantum_state import QuantumState

# This dictionary contains the qrisp operations as values and their names as keys.
# This is advantageous because during the simulation, we don't want to synthesize them
# every time they are required
# from qrisp.circuit import op_list
# op_dict = {op().name : op for op in op_list}


# This functions determines the quantum state after executing a quantum circuit
# and afterwards extracts the probability of measuring certain bit strings
def run(qc, shots, token="", iqs=None, insert_reset=True):
    
    progress_bar = tqdm(
        desc=f"Simulating {len(qc.qubits)} qubits..",
        bar_format="{desc} |{bar}| [{percentage:3.0f}%]",
        ncols=85,
        # ascii = " >#",
        leave=False,
        delay=0.2,
        position=0,
        smoothing=1,
        file=sys.stdout
        # colour = "green"
    )
    
    progress_bar.display()

    # This command enables fast appending. Fast appending means that the .append method
    # of the QuantumCircuit class checks much less validity conditions and is also less
    # tolerant regarding inputs.
    with fast_append():
        # We convert the circuit which is given with portable objects to a qrisp circuit
        qc = convert_circuit(qc, "qrisp", transpile=True)

        # Count the amount of measurements (we can stop the simulation after all
        # measurements are performed)
        measurement_amount = count_measurements_and_treat_alloc(
            qc, insert_reset=insert_reset
        )

        # Apply circuit preprocessing more
        qc = circuit_preprocessor(qc)

        measurement_counter = 0

        for i in range(len(qc.data)):
            if qc.data[i].op.name == "measure":
                measurement_counter += 1
            if measurement_counter == measurement_amount:
                break

        qc.data = qc.data[: i + 1]

        if iqs is None:
            # Create impure quantum state object. This object tracks multiple decoherent
            # quantum states that can appear when applying a non-unitary operation
            iqs = ImpureQuantumState(len(qc.qubits), clbit_amount=len(qc.clbits))

        # Counter to track how many measurements have been performed
        measurement_counter = 0

        total_flops = 0
        for i in range(len(qc.data)):
            total_flops += 2 ** qc.data[i].op.num_qubits

        progress_bar.total = total_flops
        # Main loop - this loop successively executes operations onto the impure
        # quantum state object
        for i in range(len(qc.data)):
            # Set alias for the instruction of this operation
            instr = qc.data[i]

            progress_bar.update(2**instr.op.num_qubits)

            # Gather the indices of the qubits from the circuits (i.e. integers instead
            # of the identifier strings)
            qubit_indices = [qc.qubits.index(qb) for qb in instr.qubits]

            # Perform instructions

            if instr.op.name == "reset":
                iqs.reset(qubit_indices[0])

            # Disentangling describes an operation, which mean that the superposition of
            # two states can be safely treated as two decoherent states. This is
            # advantageous because it might be possible that the amplitude of one
            # state is 0, which means that we halfed the workload. Even if both states
            # have non-zero amplitude, this still yields an improvement because
            # computing two decoherent states is more easily parallelized than the
            # combined coherent state
            elif instr.op.name == "disentangle":
                iqs.reset(qubit_indices[0], True)

            elif instr.op.name == "measure":
                iqs.measure(qubit_indices[0], qc.clbits.index(instr.clbits[0]))
                measurement_counter += 1

            elif instr.op.name[:4] == "c_if":
                iqs.conditionally_apply_operation(
                    instr.op, qubit_indices, qc.clbits.index(instr.clbits[-1])
                )
            # If the operation is unitary, we apply this unitary on to the required
            # qubit indices
            else:
                iqs.apply_operation(instr.op, qubit_indices)

            # If all measurements have been performed, break
            if measurement_counter == measurement_amount:
                break

        progress_bar.close()

        LINE_UP = "\033[1A"
        LINE_CLEAR = "\x1b[2K"
        print(LINE_CLEAR, end="\r")
        # print(LINE_UP, end = "")

        # Prepare result dictionary
        # The iqs object contains the outcome bitstrings in the attribute .outcome_list
        # and the probablities in .cl_prob. In order to ensure qiskit compatibility, we
        # reverse the bitstrings
        prob_dict = {}
        for i in range(len(iqs.cl_prob)):
            if iqs.cl_prob[i] == 0:
                continue

            outcome_str = ""
            for j in range(len(iqs.outcome_list[i])):
                if iqs.outcome_list[i][j]:
                    outcome_str += "1"
                else:
                    outcome_str += "0"

            try:
                prob_dict[outcome_str[::-1]] += iqs.cl_prob[i]
            except KeyError:
                prob_dict[outcome_str[::-1]] = iqs.cl_prob[i]

        
        #Normalize probabilities to 1 (can be slightly different due to
        #floating point errors)
        norm = 0
        for v in prob_dict.values():
            norm += v
            
        for k,v in prob_dict.items():
            prob_dict[k] = prob_dict[k]/norm
            

        #If shots >= 10000, no samples will be drawn and the distribution will
        #be returned instead
        if shots >= 10000:
            res = {}
            
            for k, v in prob_dict.items():
                res[k] = int(np.round(v*abs(shots)))
                
        #Generate samples
        else:
            
            from numpy.random import choice
            p_array = np.array(list(prob_dict.values()))
            
            samples = choice(len(p_array), shots, p=p_array)
            
            keys = list(prob_dict.keys())
            
            res = {k : 0 for k in keys}
            for s in samples:
                res[keys[s]] += 1
                
        return res


def statevector_sim(qc):
    
    progress_bar = tqdm(
        desc=f"Simulating {len(qc.qubits)} qubits..",
        bar_format="{desc} |{bar}| [{percentage:3.0f}%]",
        ncols=85,
        # ascii = " >#",
        leave=False,
        delay=0.2,
        position=0,
        smoothing=1,
        file=sys.stdout
        # colour = "green"
    )
    
    # This command enables fast appending. Fast appending means that the .append method
    # of the QuantumCircuit class checks much less validity conditions and is also less
    # tolerant regarding inputs.
    with fast_append():
        # We convert the circuit which is given with portable objects to a qrisp circuit

        if not isinstance(qc, QuantumCircuit):
            qc = convert_circuit(qc, "qrisp", transpile=True)

        qc = qc.copy()

        # TO-DO fix qubit ordering
        # qc.qubits = qc.qubits[::-1]

        # Count the amount of measurements (we can stop the simulation after all
        # measurements are performed)
        measurement_amount = count_measurements_and_treat_alloc(qc, insert_reset=False)

        if measurement_amount != 0:
            raise Exception(
                "Tried to determine the statevector of a circuit containing a "
                "measurement"
            )

        # Apply circuit preprocessing more
        qc = group_qc(qc)

        if len(qc.data) == 0:
            res = np.zeros(2 ** len(qc.qubits), dtype=np.complex64)
            res[0] = 1
            return res

        qs = QuantumState(len(qc.qubits))

        total_flops = 0
        for i in range(len(qc.data)):
            total_flops += 2 ** qc.data[i].op.num_qubits

        progress_bar.total = total_flops

        # Main loop - this loop successively executes operations onto the impure
        # quantum state object
        for i in range(len(qc.data)):
            # Set alias for the instruction of this operation
            instr = qc.data[i]

            progress_bar.update(2**instr.op.num_qubits)
            # Gather the indices of the qubits from the circuits (i.e. integers instead
            # of the identifier strings)
            qubit_indices = [qc.qubits.index(qb) for qb in instr.qubits]

            # Perform instructions
            qs.apply_operation(instr.op, qubit_indices)

        res = qs.eval().tensor_array.to_array()

        progress_bar.close()

        LINE_UP = "\033[1A"
        LINE_CLEAR = "\x1b[2K"
        print(LINE_CLEAR, end="\r")
        # print(LINE_UP, end = "")

        # Deactivate the fast append mode
        QuantumCircuit.fast_append = False

        return res


def single_shot_sim(qc, quantum_state=None):
    if len(qc.data) == 0:
        return "", quantum_state

    # This command enables fast appending. Fast appending means that the .append method
    # of the QuantumCircuit class checks much less validity conditions and is also less
    # tolerant regarding inputs.
    with fast_append():
        # We convert the circuit which is given with portable objects to a qrisp circuit
        qc = convert_circuit(qc, "qrisp", transpile=True)

        # Treat allocation gates (ie. remove them)
        count_measurements_and_treat_alloc(qc, insert_reset=True)

        from qrisp.simulator import reorder_circuit

        qc = group_qc(qc)

        qc = reorder_circuit(qc, ["measure", "reset", "disentangle"])

        if quantum_state is None:
            # Create quantum state object.
            quantum_state = QuantumState(len(qc.qubits))

        result_str = len(qc.clbits) * ["0"]

        import random

        if len(qc.data):
            # Wrapper to pre-calculate the unitaries from the preprocessed circuit in
            # parallel with the main thread
            def pre_calc_unitaries(op):
                try:
                    op.get_unitary()
                except:
                    pass

            # pre_calc_unitaries()

            pre_calc_thr = threading.Thread(
                target=pre_calc_unitaries, args=(qc.data[0].op,)
            )
            pre_calc_thr.start()

        # Main loop - this loop successively executes operations onto the impure
        # quantum state object
        for i in range(len(qc.data)):
            pre_calc_thr.join()

            if i < len(qc.data) - 1:
                pre_calc_thr = threading.Thread(
                    target=pre_calc_unitaries, args=(qc.data[i + 1].op,)
                )
                pre_calc_thr.start()

            # Set alias for the instruction of this operation
            instr = qc.data[i]

            # Gather the indices of the qubits from the circuits (i.e. integers instead
            # of the identifier strings)
            qubit_indices = [qc.qubits.index(qb) for qb in instr.qubits]

            # Perform instructions
            if instr.op.name == "reset":
                quantum_state.measure(qubit_indices[0])

                p_0, state_0, p_1, state_1 = quantum_state.last_mes_outcome

                rng = random.random()

                if rng < p_0:
                    quantum_state = state_0
                else:
                    quantum_state = state_1

            # Disentangling describes an operation, which mean that the superposition of
            # two states can be safely treated as two decoherent states. This is
            # advantageous because it might be possible that the amplitude of one state
            # is 0, which means that we halfed the workload. Even if both states have
            # non-zero amplitude, this still yields an improvement because computing two
            # decoherent states is more easily parallelized than the combined coherent
            # state.

            elif instr.op.name == "measure":
                quantum_state.measure(qubit_indices[0])

                p_0, state_0, p_1, state_1 = quantum_state.last_mes_outcome

                rng = random.random()

                if rng < p_0:
                    quantum_state = state_0
                else:
                    quantum_state = state_1
                    result_str[qc.clbits.index(instr.clbits[0])] = "1"

            elif instr.op.name[:4] == "c_if":
                if result_str[qc.clbits.index(instr.clbits[0])] == "1":
                    quantum_state.apply_operation(instr.op, qubit_indices)
            # If the operation is unitary, we apply this unitary on to the required qubit
            # indices
            else:
                quantum_state.apply_operation(instr.op, qubit_indices)

        return "".join(result_str)[::-1], quantum_state



"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""

# Created by ann81984 at 06.05.2022
from qrisp.arithmetic import QuantumFloat, q_divmod
from qrisp.misc import multi_measurement


def test_quantum_divison():
    numerator = QuantumFloat(3, -2, signed=True)
    divisor = QuantumFloat(3, -2, signed=True)

    n = 4 / 8
    d = -5 / 4
    numerator.encode(n)
    divisor.encode(d)

    prec = 4
    quotient, remainder = q_divmod(numerator, divisor, prec=prec, adder="thapliyal")

    q, r = list(multi_measurement([quotient, remainder]))[0]

    print("Q: ", q)
    print("delta_Q: ", abs(q - n / d))
    print("R: ", r)
    assert q == -0.375 and r == 0.03125 and abs(q - n / d) < 2**-prec



"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""

# -*- coding: utf-8 -*-

from qrisp.simulator.numerics_config import xp
from qrisp.simulator.tensor_factor import TensorFactor, multi_entangle

# This class describes a quantum state
# This is achieved by another class, the TensorFactor.
# Each qubit initially has its own tensor factor. When an entangling operation is
# executed, the TensorFactors can be entangled. We then have the situation that two
# qubits are described by the same TensorFactor


# This structure allows the simulation to perform entangling operations only when they
# are really needed. If a circuit consists of two blocks which are never entangled,
# after the simulation, the QuantumState will consist of two TensorFactors.
class QuantumState:
    def __init__(self, n):
        # The index of the tensor factors refers to what qubit they are describing
        self.tensor_factors = [TensorFactor([i]) for i in range(n)]
        self.n = n

    # This functions applies a (unitary) operation onto the tensor factor
    def apply_operation(self, operation, qubits):
        # Determine the involved tensor factors and the involved qubits
        involved_factors = list(
            set(sum([[self.tensor_factors[qubits[i]]] for i in range(len(qubits))], []))
        )
        involved_qubits = list(set(sum([tf.qubits for tf in involved_factors], [])))

        # Entangle all the factors involved in the operation
        entangled_factor = multi_entangle(involved_factors)

        # Update the entangled factor on the involved qubits
        for i in range(len(involved_qubits)):
            self.tensor_factors[involved_qubits[i]] = entangled_factor

        # Get the unitary of the operation
        unitary = operation.get_unitary()
        # unitary = operation.get

        # Apply the matrix
        entangled_factor.apply_matrix(unitary, qubits)

    # Method to retrieve the statevector in the form of an array of complex numbers
    def eval(self):
        [factor.unravel() for factor in self.tensor_factors]

        # print(len(list(set(self.tensor_factors))))

        entangled_factor = multi_entangle(list(set(self.tensor_factors)))

        entangled_factor.unravel()
        return entangled_factor

    # Copying method
    def copy(self):
        factor_list = list(set(self.tensor_factors))
        new_factor_list = [factor.copy() for factor in factor_list]

        res = QuantumState(self.n)

        for i in range(len(new_factor_list)):
            factor_qubits = new_factor_list[i].qubits
            for j in range(len(factor_qubits)):
                res.tensor_factors[factor_qubits[j]] = new_factor_list[i]

        return res

    # Method to perform a measurement on this quantum state
    def measure(self, i, keep_res=True):
        # Determine the TensorFactor to be measured
        tensor_factor = self.tensor_factors[i]

        # Perform measurement on this TensorFactor
        # p_0 is the measurement probability of measuring p_0,
        # outcome_tensor_f_0 is the TensorFactor that is left after a measurement with
        # outcome 0 has been performed, i.e. if the state descibes
        # |psi> = 2**-0.5 |0>|a> + 2**-0.5|1>|b>
        # we have p_0 = 0.5 and outcome_tensor_f_0 = |a>
        p_0, outcome_tensor_f_0, p_1, outcome_tensor_f_1 = tensor_factor.measure(i)

        # This code block handles several cases that mainly serve to prevent simulation
        # of QuantumStates with a probablity of 0 to be measured.
        if p_0 != 0:
            # If both measurement results have non-zero probability, we copy self
            # In case only one state has non-zero probability we can continue using self
            if p_1 != 0:
                outcome_state_0 = self.copy()
            else:
                outcome_state_0 = self

            # Update the tensor factors of the qubits that have been entangled to the
            # measured qubit
            outcome_qubits = outcome_tensor_f_0.qubits
            for j in range(len(outcome_qubits)):
                outcome_state_0.tensor_factors[outcome_qubits[j]] = outcome_tensor_f_0

            # Update the qubit that has been measured (now it's own tensor factor again)
            outcome_state_0.tensor_factors[i] = TensorFactor([i])

        else:
            outcome_state_0 = None

        # In case the probability to measure a 1 is non-zero, we can basically proceed
        # as above, but we need to set the measured qubit to the |1> state if this is
        # required
        if p_1 != 0:
            outcome_state_1 = self
            temp = outcome_tensor_f_1.qubits
            for j in range(len(temp)):
                outcome_state_1.tensor_factors[temp[j]] = outcome_tensor_f_1

            if keep_res:
                # This set the measured qubit to the |1> state (described by the
                # array [0,1] instead of [1,0] which descibes the |0> state)
                outcome_state_1.tensor_factors[i] = TensorFactor(
                    [i], xp.array([0, 1], dtype=xp.complex64)
                )
            else:
                outcome_state_1.tensor_factors[i] = TensorFactor([i])
        else:
            outcome_state_1 = None

        # log measurement outcome. Note that we can't return here, because this function
        # is called as a thread, implying it does not keep the returned result
        self.last_mes_outcome = (p_0, outcome_state_0, p_1, outcome_state_1)

        return self.last_mes_outcome

    def add_qubit(self):
        self.tensor_factors.append(TensorFactor([self.n]))
        self.n += 1



"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""


# Abstract class to describe environments
# The idea is to start dumping the circuit data into a container
# (the .env_data attribute) once the environment is entered.
# Meanwhile, the original data of the session is stored in the .original_data attribute.
# The dumping is ended when the environment is left or a child environment is entered.
# The child environment will append itself into the container, once it is left.
# As soon as the most outer environment is left all the circuit data will be compiled.
# The compilation process is specified by the inheritor of this class.
# For instance conditional environments will turn every quantum operation
# that took place in that environment into it's controlled version.

# Another level of complexity is introduced by handling the quantum sessions.
# The problem here is that we want (almost) all quantum sessions,
# that operate inside the environment, to be merged into the environment session.
# The sessions we dont want merged are the ones that get CREATED inside the environment.
# This situtation represents the case that we have a foreign function from a module,
# which creates some gate objects, using a quantum session. If we merge this quantum
# session into our environment, the foreign function no longer works as intended.
# The strategy for solving this kind of problem is now:

# 1. At environment entry, all currently live quantum sessions are logged and the
# environments very own QuantumSession (.env_qs) is created. This QuantumEnvironment
# appends itself to the environment stack of these QuantumSessions.

# 2. Once the append method of a QuantumSession is executed, the environment stack of
# this QuantumSession is checked for Environment where the env_qs QuantumSession
# is not merged into. If this is the case, the QuantumSession's data will be transfered
# into the original_data attribute of the oldest env_qs environment
# that has not been merged into self.

# 3. At environment exit, every quantum session, that operated inside this environment
# is merged together.

# 4. Apart from the (de)allocation gates, all the collected data is stored inside
# the .env_data attribute

from qrisp.circuit import QubitAlloc, QubitDealloc
from qrisp.core.quantum_session import QuantumSession


class QuantumEnvironment:
    """

    QuantumEnvironments are blocks of code, that undergo some user-specified compilation
    process. They can be entered using the ``with`` statement

    ::

       from qrisp import QuantumEnvironment, QuantumVariable, x

       qv = QuantumVariable(5)

       with QuantumEnvironment():
          x(qv)

    In this case we have no special compilation technique, since the abstract baseclass
    simply returns it's content:

    >>> print(qv.qs)
    QuantumCircuit:
    ---------------
          ┌───┐
    qv.0: ┤ X ├
          ├───┤
    qv.1: ┤ X ├
          ├───┤
    qv.2: ┤ X ├
          ├───┤
    qv.3: ┤ X ├
          ├───┤
    qv.4: ┤ X ├
          └───┘
    Live QuantumVariables:
    ----------------------
    QuantumVariable qv

    More advanced environments allow for a large variety of features and can significantly
    simplify code development and maintainance.

    The most important built-in QuantumEnvironments are:

    * :ref:`ConditionEnvironment`

    * :ref:`ControlEnvironment`

    * :ref:`InversionEnvironment`

    * :ref:`GateWrapEnvironment`

    Due to sophisticated condition evaluation of nested :ref:`conditionenvironment` and
    :ref:`controlenvironment`, using QuantumEnvironments even can bring an increase in
    performance, compared to the `control method
    <https://qiskit.org/documentation/stubs/qiskit.circuit.QuantumCircuit.control.html>`_
    which is commonly implemented by QuantumCircuit-based approaches.

    **Uncomputation within QuantumEnvironments**

    Uncomputation via the :meth:`uncompute <qrisp.QuantumVariable.uncompute>` method is
    possible only if the :ref:`QuantumVariable` has been created within the same or a
    sub-environment:

    ::

        from qrisp import QuantumVariable, QuantumEnvironment, cx

        a = QuantumVariable(1)

        with QuantumEnvironment():

            b = QuantumVariable(1)

            cx(a,b)

            with QuantumEnvironment():

                c = QuantumVariable(1)

                cx(b,c)

            c.uncompute() # works because c was created in a sub environment
            b.uncompute() # works because b was created in the same environment
            # a.uncompute() # doesn't work because a was created outside this
            environment.


    >>> print(a.qs)
    QuantumCircuit:
    ---------------
    a.0: ──■──────────────■──
         ┌─┴─┐          ┌─┴─┐
    b.0: ┤ X ├──■────■──┤ X ├
         └───┘┌─┴─┐┌─┴─┐└───┘
    c.0: ─────┤ X ├┤ X ├─────
              └───┘└───┘
    Live QuantumVariables:
    ----------------------
    QuantumVariable a

    **Visualisation within QuantumEnvironments**

    Calling ``print`` on a :ref:`QuantumSession` inside a QuantumEnvironment will
    display only the instructions, that have been performed within this environment. ::

        from qrisp import x, y, z
        a = QuantumVariable(3)

        x(a[0])

        with QuantumEnvironment():

            y(a[1])

            with QuantumEnvironment():

                z(a[2])

                print(a.qs)

            print(a.qs)

        print(a.qs)

    Executing this snippet yields

    >>>
    QuantumCircuit:
    ---------------
    a.0: ─────
    <BLANKLINE>
    a.1: ─────
         ┌───┐
    a.2: ┤ Z ├
         └───┘
    QuantumEnvironment Stack:
    -------------------------
    Level 0: QuantumEnvironment
    Level 1: QuantumEnvironment
    <BLANKLINE>
    Live QuantumVariables:
    ----------------------
    QuantumVariable a
    QuantumCircuit:
    ---------------
    a.0: ─────
         ┌───┐
    a.1: ┤ Y ├
         ├───┤
    a.2: ┤ Z ├
         └───┘
    QuantumEnvironment Stack:
    -------------------------
    Level 0: QuantumEnvironment
    <BLANKLINE>
    Live QuantumVariables:
    ----------------------
    QuantumVariable a
    QuantumCircuit:
    ---------------
         ┌───┐
    a.0: ┤ X ├
         ├───┤
    a.1: ┤ Y ├
         ├───┤
    a.2: ┤ Z ├
         └───┘
    Live QuantumVariables:
    ----------------------
    QuantumVariable a


    .. warning::

        Calling ``print`` within a QuantumEnvironment causes all sub environments to be
        compiled. While this doesn't change the semantics of the resulting circuit,
        especially nested :ref:`Condition <conditionenvironment>`- and
        :ref:`ControlEnvironments <controlenvironment>` lose a lot of efficiency if
        compiled prematurely. Therefore, ``print``-calls within QuantumEnvironments are
        usefull for debugging purposes but should be removed, if efficiency is a
        concern.


    **Creating custom QuantumEnvironments**

    More interesting QuantumEnvironments can be created by inheriting and modifying
    the compile method. In the following code snippet, we will demonstrate how to
    set up a QuantumEnvironment, that skips every second instruction. We do this
    by inheriting from the QuantumEnvironment class. This will provide us with
    the necessary attributes for writing the compile method:

    #. ``.env_data``, which is the list of instructions, that have been appended in this
    environment. Note that child environments append themselves in this list upon
    exiting.

    #. ``.env_qs`` which is a QuantumSession, where all QuantumVariables, that operated
    inside this environment, are registered.

    The compile method is then called once all environments of ``.env_qs`` have been
    exited. Note that this doesn't neccessarily imply that all QuantumEnvironments have
    been left. For more information about the interplay between QuantumSessions and
    QuantumEnvironments check the :ref:`session merging <SessionMerging>` documentation.

    ::

       class ExampleEnvironment(QuantumEnvironment):

          def compile(self):

             for i in range(len(self.env_data)):

                #This line makes sure every second instruction is skipped
                if i%2:
                   continue

                instruction = self.env_data[i]

                #If the instruction is an environment, we compile this environment
                if isinstance(instruction, QuantumEnvironment):
                   instruction.compile()
                #Otherwise we append
                else:
                    self.env_qs.append(instruction)

    Check the result: ::

       from qrisp import x, y, z, t, s, h
       qv = QuantumVariable(6)

       with ExampleEnvironment():
           x(qv[0])
           y(qv[1])
           with ExampleEnvironment():
               z(qv[2])
               t(qv[3])
           with ExampleEnvironment():
               s(qv[4])
           h(qv[5])

    >>> print(qv.qs)
    QuantumCircuit:
    ---------------
          ┌───┐
    qv.0: ┤ X ├
          └───┘
    qv.1: ─────
          ┌───┐
    qv.2: ┤ Z ├
          └───┘
    qv.3: ─────
    <BLANKLINE>
    qv.4: ─────
          ┌───┐
    qv.5: ┤ H ├
          └───┘
    Live QuantumVariables:
    ----------------------
    QuantumVariable qv

    """

    deepest_environment = [None]

    # The methods to start the dumping process for this environment
    # The dumping basically consists of copying the original data into a temporary
    # container (here the list .original_data) and then clearing the data of
    # the quantum session. Once the dumping ends the data which has been appended
    # in the meantime is appended to the environments' data list .env_data and
    # the original circuit data is reinstated
    def start_dumping(self):
        qs = self.env_qs

        # Temporarily store the qs circuit data
        self.original_data += qs.data

        # Clear the qs circuit data to collect what is coming
        qs.clear_data()

        qs.data.extend(self.env_data)

        self.env_data = []

    def stop_dumping(self):
        qs = self.env_qs

        # Collect circuit data into the data_list
        self.env_data += qs.data

        qs.clear_data()

        # Reinstate original circuit data
        qs.data.extend(self.original_data)

        self.original_data = []

    # Method to enter the environment
    def __enter__(self):
        # The QuantumSessions operating inside this environment will be merged
        # into this QuantumSession
        self.env_qs = QuantumSession()

        # This list stores the original data of the quantum session tracked
        self.original_data = []

        # This list stores the data that is appended inside the environemt
        self.env_data = []

        # Set the new relationships
        self.parent = self.deepest_environment[0]
        self.deepest_environment[0] = self

        # Acquire a list of all active quantum sessions
        self.active_qs_list = QuantumSession.get_active_quantum_sessions()
        for qs in self.active_qs_list:
            # Append self to the environment stack of the QuantumSession
            qs().env_stack.append(self)

        # Start the dumping process
        self.start_dumping()

        return self

    # Method to exit the environment
    def __exit__(self, exception_type, exception_value, traceback):
        self.deepest_environment[0] = self.parent

        # Stop dumping
        self.stop_dumping()

        for i in range(len(self.active_qs_list)):
            # Remove from the environment stack
            if self.active_qs_list[i]() is not None:
                self.active_qs_list[i]().env_stack.pop(-1)

        if not hasattr(self, "manual_allocation_management"):
            # Create a list which will store deallocation gates
            dealloc_qubit_list = []
            alloc_qubit_list = []
            # We now iterate through the collected data. We do this in order
            # to make sure that the (de)allocation gates are notprocessed
            # by the environment compiler as this might disturb their functionality
            i = 0

            while i < len(self.env_data):
                instr = self.env_data[i]

                if not isinstance(instr, QuantumEnvironment):
                    if instr.op.name == "qb_alloc":
                        alloc_qubit_list.append(self.env_data.pop(i).qubits[0])
                        continue
                    if instr.op.name == "qb_dealloc":
                        dealloc_qubit_list.append(self.env_data.pop(i).qubits[0])
                        instr.qubits[0].allocated = True
                        continue

                i += 1

            for qb in list(set(alloc_qubit_list)):
                self.env_qs.append(QubitAlloc(), [qb])

        # If this was the outermost environment, we compile
        if len(self.env_qs.env_stack) == 0:
            self.compile()

        # Otherwise, we append self to the data of the parent environment
        else:
            if len(self.env_data):
                self.env_qs.data.append(self)

        if not hasattr(self, "manual_allocation_management"):
            for qb in list(set(dealloc_qubit_list)):
                self.env_qs.append(QubitDealloc(), [qb])

    # This is the default compilation method.
    # It simply compiles all subenvironments and the collected data to the session
    # For more advanced environments this should be modified
    def compile(self):
        for instruction in self.env_data:
            # If the instruction is an environment, compile the environment
            if issubclass(instruction.__class__, QuantumEnvironment):
                instruction.compile()

            else:
                # Append instruction
                self.env_qs.append(instruction)



"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""

# Created by ann81984 at 05.05.2022
import pytest
import numpy as np

from qrisp import QuantumSession, QuantumVariable
from qrisp.misc.GMS_tools import (
    gms_multi_cx_fan_out,
    gms_multi_cx_fan_in,
    gms_multi_cp_gate,
    gms_multi_cp_gate_mono_phase,
    GXX_wrapper,
)
from qrisp import x


def test_GXX_gates_example():
    n = 5
    qs = QuantumSession()
    qv = QuantumVariable(n, qs)

    x(qv[-1])
    qs.append(
        gms_multi_cx_fan_out(
            n - 1, use_uniform=True, phase_tolerant=False, basis="GXX"
        ),
        qv.reg,
    )

    tmp = qv.get_measurement()

    print(tmp)
    assert qv.get_measurement() == {"11111": 1.0}

    ###################
    n = 3

    theta = [np.pi / (i + 1) for i in range(n - 1)]
    qs_0 = QuantumSession()
    qv = QuantumVariable(n, qs_0)
    qs_0.append(gms_multi_cp_gate(n - 1, theta), qv.reg)

    qs_1 = QuantumSession()
    qv = QuantumVariable(n, qs_1)

    for i in range(n - 1):
        qs_1.cp(theta[i], i, n - 1)

    unitary = qs_0.get_unitary()
    print(qs_0.get_unitary(decimals=4))
    tmparray = qs_0.get_unitary(decimals=4)
    assert type(qs_0.get_unitary(decimals=4)).__module__ == "numpy"
    while i < len(tmparray):
        assert all(tmparray[i]) in [
            1.0 + 0.0j,
            -0.0 - 0.0j,
            -0.0 + 0.0j,
            0.0 + 0.0j,
            1.0 - 0.0j,
            0.0 - 0.0j,
            0.0 + 1.0j,
            -1.0 + 0.0j,
            -0.0 - 1.0j,
        ]
        assert not bool(all(tmparray[i]) == 7153)
        i += 1

    print("Is equal:", qs_0.compare_unitary(qs_1))
    assert qs_0.compare_unitary(qs_1)

    print(
        "Is diagonal:", np.linalg.norm(unitary - np.diag(np.diagonal(unitary))) < 1e-3
    )
    assert np.linalg.norm(unitary - np.diag(np.diagonal(unitary))) < 1e-3

    print(np.round(np.angle(np.diagonal(unitary)) % (2 * np.pi) / (np.pi), 3))



"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""


from qrisp.misc.utility import lifted
from qrisp.environments import adaptive_condition


def less_than_gate(a, b):
    from qrisp import QuantumBool, QuantumFloat, cx, inpl_add

    a = a.duplicate()

    if isinstance(b, QuantumFloat):
        b = b.duplicate(qs=a.qs)

    # a[:] = -5
    added_sign = False
    if not a.signed:
        a.extend(1, position=-1)
        a.mshape[1] -= 1
        a.signed = True
        added_sign = True

    added_qubits_upper = 0

    while True:
        a.extend(1, position=a.size - 1)

        if not added_sign:
            cx(a[-1], a[-2])

        added_qubits_upper += 1

        if isinstance(b, QuantumFloat):
            if not a.mshape[1] < b.mshape[1] + 1:
                break
        else:
            break

    added_qubits_lower = 0

    if isinstance(b, QuantumFloat):
        while a.mshape[0] > b.mshape[0]:
            a.extend(1, position=0)

            a.mshape -= 1
            a.exponent -= 1

            added_qubits_lower += 1

    a -= b
    # x(a)
    # inpl_add(a, b, ignore_rounding_error=True, ignore_overflow_error=False)
    # x(a)

    if added_sign:
        res = a[-1]

        a.reduce(a[-1])
        a.qs.data.pop(-1)
    else:
        res_qbl = QuantumBool()
        res = res_qbl[0]
        cx(a[-1], res)

    a += b
    # inpl_add(a, b, ignore_rounding_error=True, ignore_overflow_error=False)

    for i in range(added_qubits_upper):
        if not added_sign:
            cx(a[-1], a[-2])

        if added_sign:
            a.reduce(a[-1], verify=False)
        else:
            a.reduce(a[-2], verify=False)

    for i in range(added_qubits_lower):
        a.reduce(a[0], verify=False)
        a.mshape += 1
        a.exponent += 1

    if added_sign:
        a.mshape[1] += 1
        a.signed = False

    if isinstance(b, QuantumFloat):
        reordered_qubits = a.reg + b.reg + [res]
    else:
        reordered_qubits = a.reg + [res]
    qc = a.qs.compile(cancel_qfts=False)
    # qc = a.qs

    for qb in qc.qubits:
        if qb not in reordered_qubits:
            reordered_qubits.append(qb)

    qc.qubits = reordered_qubits

    res_gate = qc.to_gate("less_than")
    res_gate.is_qfree = True
    res_gate.permeability = {
        i: i < qc.qubits.index(res) for i in range(res_gate.num_qubits)
    }

    return res_gate

@lifted
def less_than(a, b):
    from qrisp import QuantumBool, QuantumFloat, QuantumVariable, x

    if isinstance(a, QuantumFloat) and isinstance(b, QuantumFloat):
        lt_gate = less_than_gate(a, b)

        lt_qbl = QuantumBool()

        anc_amount = lt_gate.num_qubits - a.size - b.size - 1

        if anc_amount:
            lt_ancilla = QuantumVariable(anc_amount)
            ancillae = lt_ancilla.reg
        else:
            ancillae = []

        a.qs.append(lt_gate, a.reg + b.reg + lt_qbl.reg + ancillae)

        if anc_amount:
            lt_ancilla.delete(verify=False)

        return lt_qbl

    elif isinstance(a, QuantumFloat):
        labels = [a.decoder(i) for i in range(2**a.size)]
        labels.sort()

        if labels[-1] < b:
            return always_true()
        if labels[0] >= b:
            return always_false()

        for i in range(len(labels) - 1):
            if labels[i] < b <= labels[i + 1]:
                b = labels[i + 1]
                break
        else:
            b = labels[-1]

        lt_gate = less_than_gate(a, b)

        lt_qbl = QuantumBool()

        anc_amount = lt_gate.num_qubits - a.size - 1

        if anc_amount:
            lt_ancilla = QuantumVariable(anc_amount)
            ancillae = lt_ancilla.reg
        else:
            ancillae = []

        a.qs.append(lt_gate, a.reg + lt_qbl.reg + ancillae)

        if anc_amount:
            lt_ancilla.delete(verify=False)

        return lt_qbl

    elif isinstance(b, QuantumFloat):
        added_sign = False
        if not b.signed:
            b.extend(1, b.size)
            b.mshape[1] -= 1
            b.signed = True
            added_sign = True

        x(b)
        res = less_than(b, -a - 2**b.exponent)

        x(b)

        if added_sign:
            b.reduce(b.reg[-1])
            b.mshape[1] += 1
            b.signed = False

        return res


@lifted
def equal(qf_0, qf_1):
    from qrisp import QuantumBool, QuantumFloat, cx, mcx

    eq_qbl = QuantumBool()

    if isinstance(qf_1, QuantumFloat):
        if qf_1.signed and not qf_0.signed:
            qf_0, qf_1 = qf_1, qf_0

        mcx_qubits = []

        if qf_1.signed and qf_0.signed:
            cx(qf_1.sign(), qf_0.sign())
            mcx_qubits.append(qf_0.sign())

        elif qf_0.signed:
            mcx_qubits.append(qf_0.sign())

        significance_dict = {}

        for i in range(qf_0.msize):
            significance_dict[qf_0.exponent + i] = [qf_0[i]]
            mcx_qubits.append(qf_0[i])

        for i in range(qf_1.msize):
            if i + qf_1.exponent in significance_dict:
                cx(qf_1[i], significance_dict[i + qf_1.exponent])
            else:
                mcx_qubits.append(qf_1[i])

        mcx(mcx_qubits, eq_qbl, ctrl_state=0)

        for i in range(qf_1.msize):
            if i + qf_1.exponent in significance_dict:
                cx(qf_1[i], significance_dict[i + qf_1.exponent])

        if qf_1.signed and qf_0.signed:
            cx(qf_1.sign(), qf_0.sign())

        return eq_qbl

    if qf_0.truncate(qf_1) != qf_1:
        return always_false()

    mcx(qf_0, eq_qbl, ctrl_state=qf_0.encoder(qf_1))

    return eq_qbl


def always_true():
    from qrisp import QuantumBool

    true_qbl = QuantumBool()
    true_qbl.flip()
    return true_qbl


def always_false():
    from qrisp import QuantumBool

    false_qbl = QuantumBool()
    return false_qbl


@adaptive_condition
def lt(a, b):
    return less_than(a, b)


@adaptive_condition
def gt(a, b):
    return less_than(b, a)


@adaptive_condition
def geq(a, b):
    return lt(a, b).flip()


@adaptive_condition
def leq(a, b):
    return gt(a, b).flip()


@adaptive_condition
def eq(a, b):
    return equal(a, b)


@adaptive_condition
def neq(a, b):
    return equal(a, b).flip()



"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""

# Created by ann81984 at 05.05.2022
import pytest
import numpy as np

from qrisp import QuantumSession, QuantumVariable, QuantumCircuit
from qrisp.misc.GMS_tools import (
    gms_multi_cx_fan_out,
    gms_multi_cx_fan_in,
    gms_multi_cp_gate,
    gms_multi_cp_gate_mono_phase,
    GXX_wrapper,
    GZZ_converter,
)
from qrisp import x


def test_GZZ_gates_example():
    n = 5
    # M = np.array([[0,0,1,1,1,1,0,1,0], [0,1,0,0,1,0,1,1,1],[0,1,0,1,1,1,0,0,1],[1,0,0,0,1,1,0,0,0],[0,0,1,0,1,1,0,0,1],[1,1,1,0,1,1,1,1,1],[1,1,0,0,0,1,1,0,0],[0,1,1,0,1,1,0,0,0],[0,1,0,0,0,1,1,0,1]])

    def generate_random_inv_matrix(n, bit):
        from qrisp.misc import is_inv
        import random

        found = False

        while found == False:
            matrix = np.zeros((n, n))
            for i in range(n):
                for j in range(n):
                    matrix[i, j] = random.randint(0, 2**bit - 1)

            det = np.round(np.linalg.det(matrix) % 2**bit)

            found = is_inv(det, bit)

        return matrix

    M = generate_random_inv_matrix(n, 1)
    n = M.shape[0]
    row_swaps = []

    def swap_rows(M, i, j):
        M[[i, j]] = M[[j, i]]
        row_swaps.append((i, j))

    qc = QuantumCircuit(n)
    qc_gms = QuantumCircuit(n)
    K = M.copy()

    for i in range(n):
        for j in range(i, n):
            if K[j, i]:
                if i != j:
                    swap_rows(K, j, i)
                break
        else:
            raise

        fan_ins = [i]
        for j in range(n):
            if i != j and K[i, j]:
                qc.cx(j, i)
                fan_ins.append(j)
        if len(fan_ins) > 1:
            fan_ins.reverse()
            qc_gms.append(
                gms_multi_cx_fan_in(
                    len(fan_ins) - 1,
                    use_uniform=False,
                    phase_tolerant=False,
                    basis="GZZ",
                ),
                fan_ins,
            )

        temp = K[i, :].copy()
        temp[i] = 0
        for j in range(i, n):
            if K[j, i]:
                K[j, :] = (K[j, :] + temp) % 2

    assert qc.compare_unitary(qc_gms, precision=4) == True


def test_GZZ_converter():
    qc = QuantumCircuit(7)
    qc.cz(qc.qubits[0], qc.qubits[1])
    qc.cz(qc.qubits[0], qc.qubits[3])
    qc.cz(qc.qubits[5], qc.qubits[4])
    qc.p(0.5, qc.qubits[2])
    qc.cp(0.25, qc.qubits[1], qc.qubits[6])
    qc.cz(qc.qubits[3], qc.qubits[1])
    qc.cz(qc.qubits[2], qc.qubits[6])
    qc.rz(0.5, qc.qubits[4])
    qc.cz(qc.qubits[3], qc.qubits[4])

    qc_zz = GZZ_converter(qc)

    assert qc.compare_unitary(qc_zz, 4) == True
